# ruff: noqa: UP006
# ruff: noqa: UP007
# ruff: noqa: UP045
import sys
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Protocol,
    Sequence,
    SupportsIndex,
    Tuple,
    TypeVar,
    Union,
)

if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias

from .MPI import (
    BottomType,
    Datatype,
    InPlaceType,
)

__all__: List[str] = [
    "SupportsBuffer",
    "SupportsDLPack",
    "SupportsCAI",
    "Buffer",
    "Bottom",
    "InPlace",
    "Aint",
    "Count",
    "Displ",
    "Offset",
    "TypeSpec",
    "BufSpec",
    "BufSpecB",
    "BufSpecV",
    "BufSpecW",
    "TargetSpec",
]

_Stream: TypeAlias = Union[int, Any]
_PyCapsule: TypeAlias = object
_DeviceType: TypeAlias = int
_DeviceID: TypeAlias = int

class SupportsBuffer(Protocol):
    if sys.version_info >= (3, 12):
        def __buffer__(self, flags: int, /) -> memoryview: ...

class SupportsDLPack(Protocol):
    def __dlpack__(
        self, /, *, stream: Optional[_Stream] = None
    ) -> _PyCapsule: ...
    def __dlpack_device__(
        self,
    ) -> Tuple[_DeviceType, _DeviceID]: ...

class SupportsCAI(Protocol):
    @property
    def __cuda_array_interface__(self) -> Dict[str, Any]: ...

Buffer: TypeAlias = Union[
    SupportsBuffer,
    SupportsDLPack,
    SupportsCAI,
]

Bottom: TypeAlias = Union[BottomType, None]

InPlace: TypeAlias = Union[InPlaceType, None]

Aint: TypeAlias = SupportsIndex

Count: TypeAlias = SupportsIndex

Displ: TypeAlias = SupportsIndex

Offset: TypeAlias = SupportsIndex

TypeSpec: TypeAlias = Union[Datatype, str]

BufSpec: TypeAlias = Union[
    Buffer,
    Tuple[Buffer, Count],
    Tuple[Buffer, TypeSpec],
    Tuple[Buffer, Count, TypeSpec],
    Tuple[Bottom, Count, Datatype],
    List[Any],
]

BufSpecB: TypeAlias = Union[
    Buffer,
    Tuple[Buffer, Count],
    Tuple[Buffer, TypeSpec],
    Tuple[Buffer, Count, TypeSpec],
    List[Any],
]

BufSpecV: TypeAlias = Union[
    Buffer,
    Tuple[Buffer, Sequence[Count]],
    Tuple[Buffer, Tuple[Sequence[Count], Sequence[Displ]]],
    Tuple[Buffer, TypeSpec],
    Tuple[Buffer, Sequence[Count], TypeSpec],
    Tuple[Buffer, Tuple[Sequence[Count], Sequence[Displ]], TypeSpec],
    Tuple[Buffer, Sequence[Count], Sequence[Displ], TypeSpec],
    Tuple[Bottom, Tuple[Sequence[Count], Sequence[Displ]], Datatype],
    Tuple[Bottom, Sequence[Count], Sequence[Displ], Datatype],
    List[Any],
]

BufSpecW: TypeAlias = Union[
    Tuple[Buffer, Sequence[Datatype]],
    Tuple[Buffer, Tuple[Sequence[Count], Sequence[Displ]], Sequence[Datatype]],
    Tuple[Buffer, Sequence[Count], Sequence[Displ], Sequence[Datatype]],
    Tuple[Bottom, Tuple[Sequence[Count], Sequence[Displ]], Sequence[Datatype]],
    Tuple[Bottom, Sequence[Count], Sequence[Displ], Sequence[Datatype]],
    List[Any],
]

TargetSpec: TypeAlias = Union[
    Displ,
    Tuple[()],
    Tuple[Displ],
    Tuple[Displ, Count],
    Tuple[Displ, Count, TypeSpec],
    List[Any],
]

S = TypeVar("S")  # noqa: PYI001
T = TypeVar("T")  # noqa: PYI001
U = TypeVar("U")  # noqa: PYI001
V = TypeVar("V")  # noqa: PYI001



import os
import pathlib
import shlex
import shutil

import cffi


def _read_text(filename):
    return pathlib.Path(filename).read_text(encoding="utf-8")


ffi = cffi.FFI()
ffi.set_source("libmpi", _read_text("libmpi.c.in"))
ffi.cdef(_read_text("libmpi.h"))


class mpicompiler:
    from cffi import ffiplatform

    def __init__(self, cc, ld=None):
        self.cc = cc
        self.ld = ld or cc
        self.ffi_compile = self.ffiplatform.compile

    def __enter__(self):
        self.ffiplatform.compile = self.compile

    def __exit__(self, *args):
        self.ffiplatform.compile = self.ffi_compile

    def configure(self, compiler):
        def fix_command(command, cmd):
            if not cmd:
                return
            cmd = shlex.split(cmd)
            exe = shutil.which(cmd[0])
            if not exe:
                return
            command[0] = exe
            command += cmd[1:]

        fix_command(compiler.compiler_so, self.cc)
        fix_command(compiler.linker_so, self.ld)

    def compile(self, *args, **kwargs):
        from distutils.command import build_ext

        customize_compiler_orig = build_ext.customize_compiler

        def customize_compiler(compiler):
            customize_compiler_orig(compiler)
            self.configure(compiler)

        build_ext.customize_compiler = customize_compiler
        try:
            return self.ffi_compile(*args, **kwargs)
        finally:
            build_ext.customize_compiler = customize_compiler_orig


if __name__ == "__main__":
    cc = os.environ.get("MPICC", "mpicc")
    ld = os.environ.get("MPILD")
    with mpicompiler(cc, ld):
        ffi.compile()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Convert NumPy and MPI datatypes."""
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=too-many-return-statements

from .. import MPI

try:
    from numpy import dtype as _np_dtype
except ImportError:  # pragma: no cover
    _np_dtype = None


def _get_datatype(dtype):
    return MPI.Datatype.fromcode(dtype.char)


def _get_typecode(datatype):
    # pylint: disable=protected-access
    return MPI._typecode(datatype)


def _get_alignment(datatype):
    # pylint: disable=protected-access
    return MPI._typealign(datatype)


def _is_aligned(datatype, offset=0):
    """Determine whether an MPI datatype is aligned."""
    if datatype.is_predefined:
        if offset == 0:
            return True
        alignment = _get_alignment(datatype) or 0
        return offset % alignment == 0

    combiner = datatype.combiner
    basetype, _, info = datatype.decode()
    types, disps = [basetype], [0]
    try:
        if combiner == MPI.COMBINER_RESIZED:
            disps = [info["extent"]]
        if combiner == MPI.COMBINER_STRUCT:
            types = info["datatypes"]
            disps = info["displacements"]
        if combiner == MPI.COMBINER_HVECTOR:
            disps = [info["stride"] if info["count"] > 1 else 0]
        if combiner == MPI.COMBINER_HINDEXED:
            disps = info["displacements"]
        if combiner == MPI.COMBINER_HINDEXED_BLOCK:
            disps = info["displacements"]
        return all(_is_aligned(t, offset + d) for (t, d) in zip(types, disps))
    finally:
        for _tp in types:
            if not _tp.is_predefined:
                _tp.Free()


def from_numpy_dtype(dtype):
    """Convert NumPy datatype to MPI datatype."""
    if _np_dtype is None:
        raise RuntimeError("NumPy is not available") from None
    dtype = _np_dtype(dtype)

    if dtype.hasobject:
        raise ValueError("NumPy datatype with object entries")
    if not dtype.isnative:
        raise ValueError("NumPy datatype with non-native byteorder")

    # struct data type
    fields = dtype.fields
    if fields:
        blocklengths = []
        displacements = []
        datatypes = []
        try:
            for name in dtype.names or ():
                ftype, fdisp, *_ = fields[name]
                blocklengths.append(1)
                displacements.append(fdisp)
                datatypes.append(from_numpy_dtype(ftype))
            datatype = MPI.Datatype.Create_struct(
                blocklengths, displacements, datatypes
            )
        finally:
            for mtp in datatypes:
                mtp.Free()
        try:
            return datatype.Create_resized(0, dtype.itemsize)
        finally:
            datatype.Free()

    # subarray data type
    subdtype = dtype.subdtype
    if subdtype:
        base, shape = subdtype
        datatype = from_numpy_dtype(base)
        try:
            if len(shape) == 1:
                return datatype.Create_contiguous(shape[0])
            starts = (0,) * len(shape)
            return datatype.Create_subarray(shape, shape, starts)
        finally:
            datatype.Free()

    # elementary data type
    datatype = _get_datatype(dtype)
    return datatype.Dup()


def to_numpy_dtype(datatype):
    """Convert MPI datatype to NumPy datatype."""

    def mpi2npy(datatype, count):
        dtype = to_numpy_dtype(datatype)
        return dtype if count == 1 else (dtype, count)

    def np_dtype(spec, **kwargs):
        if _np_dtype is None:
            return spec if not kwargs else (spec, kwargs)
        return _np_dtype(spec, **kwargs)

    if datatype == MPI.DATATYPE_NULL:
        raise ValueError("cannot convert null MPI datatype to NumPy")

    combiner = datatype.combiner

    # named elementary datatype
    if combiner == MPI.COMBINER_NAMED:
        # elementary datatype
        typecode = _get_typecode(datatype)
        if typecode is not None:
            return np_dtype(typecode)
        # pair datatype for MINLOC/MAXLOC reductions
        names = ("SHORT", "INT", "LONG", "FLOAT", "DOUBLE", "LONG_DOUBLE")
        types = [getattr(MPI, f"{name}_INT") for name in names]
        typename = names[types.index(datatype)]
        typecode = _get_typecode(getattr(MPI, typename))
        return np_dtype(f"{typecode},i", align=True)

    # user-defined datatype
    basetype, _, info = datatype.decode()
    datatypes = [basetype]
    try:
        # duplicated datatype
        if combiner == MPI.COMBINER_DUP:
            return to_numpy_dtype(basetype)

        # contiguous datatype
        if combiner == MPI.COMBINER_CONTIGUOUS:
            dtype = to_numpy_dtype(basetype)
            count = info["count"]
            return np_dtype((dtype, (count,)))

        # subarray datatype
        if combiner == MPI.COMBINER_SUBARRAY:
            dtype = to_numpy_dtype(basetype)
            sizes = info["sizes"]
            subsizes = info["subsizes"]
            starts = info["starts"]
            order = info["order"]
            if subsizes == sizes and min(starts) == max(starts) == 0:
                if order == MPI.ORDER_FORTRAN:
                    sizes = sizes[::-1]
                return np_dtype((dtype, tuple(sizes)))
            raise ValueError("cannot convert subarray MPI datatype to NumPy")

        # value-index datatype
        if combiner == MPI.COMBINER_VALUE_INDEX:  # pragma: no cover
            value = to_numpy_dtype(info["value"])
            index = to_numpy_dtype(info["index"])
            datatypes = [value, index]
            return np_dtype(
                [
                    ("f0", value),
                    ("f1", index),
                ],
                align=True,
            )

        # struct datatype
        aligned = True
        if combiner == MPI.COMBINER_RESIZED:
            if basetype.combiner == MPI.COMBINER_STRUCT:
                aligned = _is_aligned(basetype, info["extent"])
                combiner = MPI.COMBINER_STRUCT
                _, _, info = basetype.decode()
                datatypes.pop().Free()
        if combiner == MPI.COMBINER_STRUCT:
            datatypes = info["datatypes"]
            blocklengths = info["blocklengths"]
            displacements = info["displacements"]
            names = [f"f{i}" for i in range(len(datatypes))]
            formats = list(map(mpi2npy, datatypes, blocklengths))
            offsets = displacements
            itemsize = datatype.extent
            aligned &= all(map(_is_aligned, datatypes, offsets))
            return np_dtype({
                "names": names,
                "formats": formats,
                "offsets": offsets,
                "itemsize": itemsize,
                "aligned": aligned,
            })

        # vector datatype
        combiner_vector = (
            MPI.COMBINER_VECTOR,
            MPI.COMBINER_HVECTOR,
        )
        if combiner in combiner_vector:
            dtype = to_numpy_dtype(basetype)
            count = info["count"]
            blocklength = info["blocklength"]
            stride = info["stride"]
            if combiner == MPI.COMBINER_VECTOR:
                stride *= basetype.extent
                aligned = _is_aligned(basetype)
            if combiner == MPI.COMBINER_HVECTOR:
                stride = stride if count > 1 else 0
                aligned = _is_aligned(basetype, stride)
            names = [f"f{i}" for i in range(count)]
            formats = [(dtype, (blocklength,))] * count
            offsets = [stride * i for i in range(count)]
            itemsize = datatype.extent
            return np_dtype({
                "names": names,
                "formats": formats,
                "offsets": offsets,
                "itemsize": itemsize,
                "aligned": aligned,
            })

        # indexed datatype
        combiner_indexed = (
            MPI.COMBINER_INDEXED,
            MPI.COMBINER_HINDEXED,
            MPI.COMBINER_INDEXED_BLOCK,
            MPI.COMBINER_HINDEXED_BLOCK,
        )
        if combiner in combiner_indexed:
            dtype = to_numpy_dtype(basetype)
            stride = 1
            aligned = _is_aligned(basetype)
            blocklengths = []
            displacements = info["displacements"]
            if combiner in combiner_indexed[:2]:
                blocklengths = info["blocklengths"]
            if combiner in combiner_indexed[2:]:
                blocklengths = [info["blocklength"]] * len(displacements)
            if combiner in combiner_indexed[0::2]:
                stride = basetype.extent
            if combiner in combiner_indexed[1::2]:
                aligned &= all(_is_aligned(basetype, d) for d in displacements)
            names = [f"f{i}" for i in range(len(displacements))]
            formats = [(dtype, (blen,)) for blen in blocklengths]
            offsets = [disp * stride for disp in displacements]
            return np_dtype({
                "names": names,
                "formats": formats,
                "offsets": offsets,
                "aligned": aligned,
            })

        # Fortran 90 datatype
        combiner_f90 = (
            MPI.COMBINER_F90_INTEGER,
            MPI.COMBINER_F90_REAL,
            MPI.COMBINER_F90_COMPLEX,
        )
        if combiner in combiner_f90:
            datatypes.pop()
            typecode = _get_typecode(datatype)
            return np_dtype(typecode)

        raise ValueError("cannot convert MPI datatype to NumPy")
    finally:
        for _tp in datatypes:
            if not _tp.is_predefined:
                _tp.Free()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Pickle-based communication using protocol 5."""

import struct as _struct

from .. import MPI
from ..MPI import (
    ANY_SOURCE,
    ANY_TAG,
    PROC_NULL,
    ROOT,
    Pickle,
    Status,
    _comm_lock,
    _commctx_inter,
    _commctx_intra,
)

pickle = Pickle()


def _pickle_dumps(obj):
    return pickle.dumps_oob(obj)


def _pickle_loads(data, bufs):
    return pickle.loads_oob(data, bufs)


class _BigMPI:
    """Support for large message counts."""

    blocksize = (
        (1 << (_struct.calcsize("q") * 8 - 1)) - 1
        if MPI.Get_version() >= (4, 0)
        else (1 << (_struct.calcsize("i") * 8 - 1)) - 1
    )

    def __init__(self):
        self.cache = {}

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        for dtype in self.cache.values():
            dtype.Free()
        self.cache.clear()

    def __call__(self, buf):
        buf = memoryview(buf)
        count = buf.nbytes
        if count <= self.blocksize:
            return (buf, count, MPI.BYTE)
        dtype = self.cache.get(count)
        if dtype is None:
            dtype = MPI.BYTE.Create_contiguous(count)
            self.cache[count] = dtype.Commit()
        return (buf, 1, dtype)


_bigmpi = _BigMPI()


def _info_typecode():
    return "q"


def _info_datatype():
    code = _info_typecode()
    return MPI.Datatype.fromcode(code)


def _info_pack(info):
    code = _info_typecode()
    size = len(info)
    return _struct.pack(f"{size}{code}", *info)


def _info_alloc(size):
    code = _info_typecode()
    itemsize = _struct.calcsize(code)
    return bytearray(size * itemsize)


def _info_unpack(info):
    code = _info_typecode()
    itemsize = _struct.calcsize(code)
    size = len(info) // itemsize
    return _struct.unpack(f"{size}{code}", info)


def _new_buffer(size):
    return MPI.buffer.allocate(size)


def _send_raw(comm, send, data, bufs, dest, tag):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    info = [len(data)]
    info.extend(len(sbuf) for sbuf in bufs)
    infotype = _info_datatype()
    info = _info_pack(info)
    send(comm, (info, infotype), dest, tag)
    with _bigmpi as bigmpi:
        send(comm, bigmpi(data), dest, tag)
        for sbuf in bufs:
            send(comm, bigmpi(sbuf), dest, tag)


def _send(comm, send, obj, dest, tag):
    if dest == PROC_NULL:
        send(comm, (None, 0, MPI.BYTE), dest, tag)
        return
    data, bufs = _pickle_dumps(obj)
    with _comm_lock(comm, "send"):
        _send_raw(comm, send, data, bufs, dest, tag)


def _isend(comm, isend, obj, dest, tag):
    def send(comm, buf, dest, tag):
        sreqs.append(isend(comm, buf, dest, tag))

    sreqs = []
    _send(comm, send, obj, dest, tag)
    request = Request(sreqs)
    return request


def _recv_raw(comm, recv, buf, source, tag, status=None):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    if status is None:
        status = Status()
    MPI.Comm.Probe(comm, source, tag, status)
    source = status.Get_source()
    tag = status.Get_tag()
    infotype = _info_datatype()
    infosize = status.Get_elements(infotype)
    info = _info_alloc(infosize)
    MPI.Comm.Recv(comm, (info, infotype), source, tag, status)
    info = _info_unpack(info)
    if buf is not None:
        buf = memoryview(buf).cast("B")
        if len(buf) > info[0]:
            buf = buf[: info[0]]
        if len(buf) < info[0]:
            buf = None
    data = _new_buffer(info[0]) if buf is None else buf
    bufs = list(map(_new_buffer, info[1:]))
    with _bigmpi as bigmpi:
        recv(comm, bigmpi(data), source, tag)
        for rbuf in bufs:
            recv(comm, bigmpi(rbuf), source, tag)
    status.Set_elements(MPI.BYTE, sum(info))
    return data, bufs


def _recv(comm, recv, buf, source, tag, status):
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    if source == PROC_NULL:
        recv(comm, (None, 0, MPI.BYTE), source, tag, status)
        return None
    with _comm_lock(comm, "recv"):
        data, bufs = _recv_raw(comm, recv, buf, source, tag, status)
    return _pickle_loads(data, bufs)


def _mprobe(comm, mprobe, source, tag, status):
    if source == PROC_NULL:
        rmsg = MPI.Comm.Mprobe(comm, source, tag, status)
        return Message([rmsg])
    if status is None:
        status = Status()
    with _comm_lock(comm, "recv"):
        message = []
        numbytes = 0
        rmsg = mprobe(comm, source, tag, status)
        if rmsg is None:
            return None
        message.append(rmsg)
        source = status.Get_source()
        tag = status.Get_tag()
        infotype = _info_datatype()
        infosize = status.Get_elements(infotype)
        for _ in range(infosize):
            rmsg = MPI.Comm.Mprobe(comm, source, tag, status)
            message.append(rmsg)
            numbytes += status.Get_elements(MPI.BYTE)
        status.Set_elements(MPI.BYTE, numbytes)
        return Message(message)


def _mrecv_info(rmsg, size, status=None):
    mrecv = MPI.Message.Recv
    infotype = _info_datatype()
    info = _info_alloc(size)
    mrecv(rmsg, (info, infotype), status)
    info = _info_unpack(info)
    return info


def _mrecv_none(rmsg, mrecv, status):
    _mrecv_info(rmsg, 0, status)
    noproc = MPI.MESSAGE_NO_PROC
    mrecv(noproc, (None, 0, MPI.BYTE))
    data, bufs = _pickle_dumps(None)
    return (bytearray(data), bufs)


def _mrecv_data(message, mrecv, status=None):
    if message[0] == MPI.MESSAGE_NO_PROC:
        rmsg = message[0]
        return _mrecv_none(rmsg, mrecv, status)
    rmsg = iter(message)
    icnt = len(message) - 1
    info = _mrecv_info(next(rmsg), icnt, status)
    data = _new_buffer(info[0])
    bufs = list(map(_new_buffer, info[1:]))
    with _bigmpi as bigmpi:
        mrecv(next(rmsg), bigmpi(data))
        for rbuf in bufs:
            mrecv(next(rmsg), bigmpi(rbuf))
    if status is not None:
        status.Set_elements(MPI.BYTE, sum(info))
    return (data, bufs)


def _mrecv(message, status):
    def mrecv(rmsg, buf):
        MPI.Message.Recv(rmsg, buf)

    data, bufs = _mrecv_data(message, mrecv, status)
    return _pickle_loads(data, bufs)


def _imrecv(message):
    def mrecv(rmsg, buf):
        rreqs.append(MPI.Message.Irecv(rmsg, buf))

    rreqs = []
    data, bufs = _mrecv_data(message, mrecv)
    request = Request(rreqs)
    setattr(request, "_data_bufs", (data, bufs))  # noqa: B010
    return request


def _req_load(request):
    data_bufs = getattr(request, "_data_bufs", None)
    if request == MPI.REQUEST_NULL and data_bufs is not None:
        delattr(request, "_data_bufs")
    if data_bufs is not None:
        data, bufs = data_bufs
        obj = _pickle_loads(data, bufs)
        return obj
    return None


def _test(request, test, status):
    statuses = None if status is None else [status]
    flag = test(request, statuses)
    if flag:
        obj = _req_load(request)
        return (flag, obj)
    return (flag, None)


def _testall(requests, testall, statuses):
    if isinstance(statuses, list):
        for _ in range(len(requests) - len(statuses)):
            statuses.append(Status())
    reqarray = []
    stsarray = None
    for req in requests:
        reqarray.extend(req)
    if statuses is not None:
        stsarray = []
        for req, sts in zip(requests, statuses):
            stsarray.extend([sts] * len(req))
    flag = testall(reqarray, stsarray)
    if flag:
        objs = [_req_load(req) for req in requests]
        return (flag, objs)
    return (flag, None)


def _bcast_intra_raw(comm, bcast, data, bufs, root):
    rank = comm.Get_rank()
    if rank == root:
        info = [len(data)]
        info.extend(len(sbuf) for sbuf in bufs)
        infotype = _info_datatype()
        infosize = _info_pack([len(info)])
        bcast(comm, (infosize, infotype), root)
        info = _info_pack(info)
        bcast(comm, (info, infotype), root)
    else:
        infotype = _info_datatype()
        infosize = _info_alloc(1)
        bcast(comm, (infosize, infotype), root)
        infosize = _info_unpack(infosize)[0]
        info = _info_alloc(infosize)
        bcast(comm, (info, infotype), root)
        info = _info_unpack(info)
        data = _new_buffer(info[0])
        bufs = list(map(_new_buffer, info[1:]))
    with _bigmpi as bigmpi:
        bcast(comm, bigmpi(data), root)
        for rbuf in bufs:
            bcast(comm, bigmpi(rbuf), root)
    return data, bufs


def _bcast_intra(comm, bcast, obj, root):
    rank = comm.Get_rank()
    if rank == root:
        data, bufs = _pickle_dumps(obj)
    else:
        data, bufs = _pickle_dumps(None)
    with _comm_lock(comm, "bcast"):
        data, bufs = _bcast_intra_raw(comm, bcast, data, bufs, root)
    return _pickle_loads(data, bufs)


def _bcast_inter(comm, bcast, obj, root):
    rank = comm.Get_rank()
    size = comm.Get_remote_size()
    comm, tag, localcomm, _ = _commctx_inter(comm)
    if root == PROC_NULL:
        return None
    elif root == ROOT:
        send = MPI.Comm.Send
        data, bufs = _pickle_dumps(obj)
        _send_raw(comm, send, data, bufs, 0, tag)
        return None
    elif 0 <= root < size:
        if rank == 0:
            recv = MPI.Comm.Recv
            data, bufs = _recv_raw(comm, recv, None, root, tag)
        else:
            data, bufs = _pickle_dumps(None)
        with _comm_lock(localcomm, "bcast"):
            data, bufs = _bcast_intra_raw(localcomm, bcast, data, bufs, 0)
        return _pickle_loads(data, bufs)
    comm.Call_errhandler(MPI.ERR_ROOT)
    raise MPI.Exception(MPI.ERR_ROOT)


def _bcast(comm, bcast, obj, root):
    if comm.Is_inter():
        return _bcast_inter(comm, bcast, obj, root)
    else:
        return _bcast_intra(comm, bcast, obj, root)


def _get_p2p_backend():
    reqs = []

    def send(comm, buf, dest, tag):
        reqs.append(MPI.Comm.Isend(comm, buf, dest, tag))

    def recv(comm, buf, source, tag):
        MPI.Comm.Recv(comm, buf, source, tag)

    return reqs, send, recv


def _gather(comm, obj, root):
    reqs, send, recv = _get_p2p_backend()
    if comm.Is_inter():
        comm, tag, *_ = _commctx_inter(comm)
        size = comm.Get_remote_size()
        if root == PROC_NULL:
            send = recv = None
        elif root == MPI.ROOT:
            send = None
        elif 0 <= root < size:
            recv = None
        else:
            comm.Call_errhandler(MPI.ERR_ROOT)
            raise MPI.Exception(MPI.ERR_ROOT)
    else:
        comm, tag = _commctx_intra(comm)
        size = comm.Get_size()
        if root != comm.Get_rank():
            recv = None
        if root < 0 or root >= size:
            comm.Call_errhandler(MPI.ERR_ROOT)
            raise MPI.Exception(MPI.ERR_ROOT)

    if send:
        data, bufs = _pickle_dumps(obj)
        _send_raw(comm, send, data, bufs, root, tag)
    objs = None
    if recv:
        objs = []
        for source in range(size):
            data, bufs = _recv_raw(comm, recv, None, source, tag)
            obj = _pickle_loads(data, bufs)
            objs.append(obj)
    if send:
        MPI.Request.Waitall(reqs)
    return objs


def _scatter(comm, objs, root):
    # pylint: disable=too-many-branches
    reqs, send, recv = _get_p2p_backend()
    if comm.Is_inter():
        comm, tag, *_ = _commctx_inter(comm)
        size = comm.Get_remote_size()
        if root == PROC_NULL:
            send = recv = None
        elif root == ROOT:
            recv = None
        elif 0 <= root < size:
            send = None
        else:
            comm.Call_errhandler(MPI.ERR_ROOT)
            raise MPI.Exception(MPI.ERR_ROOT)
    else:
        comm, tag = _commctx_intra(comm)
        size = comm.Get_size()
        if root != comm.Get_rank():
            send = None
        if root < 0 or root >= size:
            comm.Call_errhandler(MPI.ERR_ROOT)
            raise MPI.Exception(MPI.ERR_ROOT)

    if send:
        if objs is None:
            objs = [None] * size
        elif not isinstance(objs, list):
            objs = list(objs)
        if len(objs) != size:
            raise ValueError(f"expecting {size} items, got {len(objs)}")
        for dest, obj in enumerate(objs):
            data, bufs = _pickle_dumps(obj)
            _send_raw(comm, send, data, bufs, dest, tag)
    obj = None
    if recv:
        data, bufs = _recv_raw(comm, recv, None, root, tag)
        obj = _pickle_loads(data, bufs)
    if send:
        MPI.Request.Waitall(reqs)
    return obj


def _allgather(comm, obj):
    reqs, send, recv = _get_p2p_backend()
    if comm.Is_inter():
        comm, tag, *_ = _commctx_inter(comm)
        size = comm.Get_remote_size()
    else:
        comm, tag = _commctx_intra(comm)
        size = comm.Get_size()

    data, bufs = _pickle_dumps(obj)
    for dest in range(size):
        _send_raw(comm, send, data, bufs, dest, tag)
    objs = []
    for source in range(size):
        data, bufs = _recv_raw(comm, recv, None, source, tag)
        obj = _pickle_loads(data, bufs)
        objs.append(obj)
    MPI.Request.Waitall(reqs)
    return objs


def _alltoall(comm, objs):
    reqs, send, recv = _get_p2p_backend()
    if comm.Is_inter():
        comm, tag, *_ = _commctx_inter(comm)
        size = comm.Get_remote_size()
    else:
        comm, tag = _commctx_intra(comm)
        size = comm.Get_size()

    if objs is None:
        objs = [None] * size
    elif not isinstance(objs, list):
        objs = list(objs)
    if len(objs) != size:
        raise ValueError(f"expecting {size} items, got {len(objs)}")
    for dest, obj in enumerate(objs):
        data, bufs = _pickle_dumps(obj)
        _send_raw(comm, send, data, bufs, dest, tag)
    objs = []
    for source in range(size):
        data, bufs = _recv_raw(comm, recv, None, source, tag)
        obj = _pickle_loads(data, bufs)
        objs.append(obj)
    MPI.Request.Waitall(reqs)
    return objs


class Request(tuple):
    """Request."""

    def __new__(cls, request=None):
        """Create and return a new object."""
        if request is None:
            request = (MPI.REQUEST_NULL,)
        if isinstance(request, MPI.Request):
            request = (request,)
        return super().__new__(cls, request)

    def __eq__(self, other):
        """Return ``self==other``."""
        if isinstance(other, Request):
            return tuple(self) == tuple(other)
        if isinstance(other, MPI.Request):
            return all(req == other for req in self)
        return NotImplemented

    def __ne__(self, other):
        """Return ``self!=other``."""
        if isinstance(other, Request):
            return tuple(self) != tuple(other)
        if isinstance(other, MPI.Request):
            return any(req != other for req in self)
        return NotImplemented

    def __bool__(self):
        """Return ``bool(self)``."""
        return any(req for req in self)

    def Free(self) -> None:
        """Free a communication request."""
        # pylint: disable=invalid-name
        for req in self:
            req.Free()

    def free(self) -> None:
        """Free a communication request."""
        for req in self:
            req.free()

    def cancel(self):
        """Cancel a communication request."""
        for req in self:
            req.Cancel()

    def get_status(self, status=None):
        """Non-destructive test for the completion of a request."""
        statuses = [status] + [None] * max(len(self) - 1, 0)
        return all(map(MPI.Request.Get_status, self, statuses))

    def test(self, status=None):
        """Test for the completion of a request."""
        return _test(self, MPI.Request.Testall, status)

    def wait(self, status=None):
        """Wait for a request to complete."""
        return _test(self, MPI.Request.Waitall, status)[1]

    @classmethod
    def get_status_all(cls, requests, statuses=None):
        """Non-destructive test for the completion of all requests."""
        arglist = [requests]
        if statuses is not None:
            ns, nr = len(statuses), len(requests)
            statuses += [Status() for _ in range(ns, nr)]
            arglist.append(statuses)
        return all(map(Request.get_status, *arglist))

    @classmethod
    def testall(cls, requests, statuses=None):
        """Test for the completion of all requests."""
        return _testall(requests, MPI.Request.Testall, statuses)

    @classmethod
    def waitall(cls, requests, statuses=None):
        """Wait for all requests to complete."""
        return _testall(requests, MPI.Request.Waitall, statuses)[1]


class Message(tuple):
    """Message."""

    def __new__(cls, message=None):
        """Create and return a new object."""
        if message is None:
            message = (MPI.MESSAGE_NULL,)
        if isinstance(message, MPI.Message):
            message = (message,)
        return super().__new__(cls, message)

    def __eq__(self, other):
        """Return ``self==other``."""
        if isinstance(other, Message):
            return tuple(self) == tuple(other)
        if isinstance(other, MPI.Message):
            return all(msg == other for msg in self)
        return NotImplemented

    def __ne__(self, other):
        """Return ``self!=other``."""
        if isinstance(other, Message):
            return tuple(self) != tuple(other)
        if isinstance(other, MPI.Message):
            return any(msg != other for msg in self)
        return NotImplemented

    def __bool__(self):
        """Return ``bool(self)``."""
        return any(msg for msg in self)

    def free(self) -> None:
        """Do nothing."""
        for msg in self:
            msg.free()

    def recv(self, status=None):
        """Blocking receive of matched message."""
        return _mrecv(self, status)

    def irecv(self):
        """Nonblocking receive of matched message."""
        return _imrecv(self)

    @classmethod
    def probe(cls, comm, source=ANY_SOURCE, tag=ANY_TAG, status=None):
        """Blocking test for a matched message."""
        return _mprobe(comm, MPI.Comm.Mprobe, source, tag, status)

    @classmethod
    def iprobe(cls, comm, source=ANY_SOURCE, tag=ANY_TAG, status=None):
        """Nonblocking test for a matched message."""
        return _mprobe(comm, MPI.Comm.Improbe, source, tag, status)


class Comm(MPI.Comm):
    """Communicator."""

    def __new__(cls, comm=None):
        """Create and return a new communicator."""
        return MPI.Comm.__new__(cls, comm)

    def send(self, obj, dest, tag=0):
        """Blocking send in standard mode."""
        _send(self, MPI.Comm.Send, obj, dest, tag)

    def bsend(self, obj, dest, tag=0):
        """Blocking send in buffered mode."""
        _send(self, MPI.Comm.Bsend, obj, dest, tag)

    def ssend(self, obj, dest, tag=0):
        """Blocking send in synchronous mode."""
        sreq = _isend(self, MPI.Comm.Issend, obj, dest, tag)
        MPI.Request.Waitall(sreq)

    def isend(self, obj, dest, tag=0):
        """Nonblocking send in standard mode."""
        return _isend(self, MPI.Comm.Isend, obj, dest, tag)

    def ibsend(self, obj, dest, tag=0):
        """Nonblocking send in buffered mode."""
        return _isend(self, MPI.Comm.Ibsend, obj, dest, tag)

    def issend(self, obj, dest, tag=0):
        """Nonblocking send in synchronous mode."""
        return _isend(self, MPI.Comm.Issend, obj, dest, tag)

    def recv(self, buf=None, source=ANY_SOURCE, tag=ANY_TAG, status=None):
        """Blocking receive."""
        return _recv(self, MPI.Comm.Recv, buf, source, tag, status)

    def irecv(self, buf=None, source=ANY_SOURCE, tag=ANY_TAG):  # noqa: ARG002
        """Nonblocking receive."""
        raise RuntimeError("unsupported")

    def sendrecv(
        self,
        sendobj,
        dest,
        sendtag=0,
        recvbuf=None,
        source=ANY_SOURCE,
        recvtag=ANY_TAG,
        status=None,
    ):
        """Send and receive."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        sreq = _isend(self, MPI.Comm.Isend, sendobj, dest, sendtag)
        robj = _recv(self, MPI.Comm.Recv, recvbuf, source, recvtag, status)
        MPI.Request.Waitall(sreq)
        return robj

    def mprobe(self, source=ANY_SOURCE, tag=ANY_TAG, status=None):
        """Blocking test for a matched message."""
        return _mprobe(self, MPI.Comm.Mprobe, source, tag, status)

    def improbe(self, source=ANY_SOURCE, tag=ANY_TAG, status=None):
        """Nonblocking test for a matched message."""
        return _mprobe(self, MPI.Comm.Improbe, source, tag, status)

    def bcast(self, obj, root=0):
        """Broadcast."""
        return _bcast(self, MPI.Comm.Bcast, obj, root)

    def gather(self, sendobj, root=0):
        """Gather."""
        return _gather(self, sendobj, root)

    def scatter(self, sendobj, root=0):
        """Scatter."""
        return _scatter(self, sendobj, root)

    def allgather(self, sendobj):
        """Gather to All."""
        return _allgather(self, sendobj)

    def alltoall(self, sendobj):
        """All to All Scatter/Gather."""
        return _alltoall(self, sendobj)


class Intracomm(Comm, MPI.Intracomm):
    """Intracommunicator."""


class Intercomm(Comm, MPI.Intercomm):
    """Intercommunicator."""



import os
import pathlib
import re
from collections import defaultdict

from coverage.files import (
    canonical_filename as _canonical_filename,
)
from coverage.plugin import (
    CoveragePlugin,
    FileReporter,
    FileTracer,
)

CYTHON_EXTENSIONS = {".pxd", ".pyx", ".pxi"}


def canonical_filename(filename):
    filename = pathlib.Path(filename).resolve()
    filename = _canonical_filename(str(filename))
    filename = pathlib.Path(filename)
    return filename


class CythonCoveragePlugin(CoveragePlugin):
    def configure(self, config):
        self.exclude = config.get_option("report:exclude_lines")

    def file_tracer(self, filename):
        filename = canonical_filename(filename)
        ext = filename.suffix
        if ext in CYTHON_EXTENSIONS:
            return CythonFileTracer(str(filename))
        return None

    def file_reporter(self, filename):
        filename = canonical_filename(filename)
        ext = filename.suffix
        if ext in CYTHON_EXTENSIONS:
            return CythonFileReporter(str(filename), self.exclude)
        return None


class CythonFileTracer(FileTracer):
    def __init__(self, filename):
        super().__init__()
        self.filename = filename

    def source_filename(self):
        return self.filename


class CythonFileReporter(FileReporter):
    def __init__(self, filename, exclude=None):
        super().__init__(filename)
        self.exclude = exclude

    def lines(self):
        _setup_lines(self.exclude)
        return self._get_lines(CODE_LINES)

    def excluded_lines(self):
        _setup_lines(self.exclude)
        return self._get_lines(EXCL_LINES)

    def translate_lines(self, lines):
        _setup_lines(self.exclude)
        exec_lines = self._get_lines(EXEC_LINES)
        return set(lines).union(exec_lines)

    def _get_lines(self, lines_map):
        key = os.path.relpath(self.filename, TOPDIR)
        lines = lines_map.get(key, {})
        return set(lines)


TOPDIR = pathlib.Path(__file__).resolve().parent.parent
SRCDIR = TOPDIR / "src"
CODE_LINES = None
EXEC_LINES = None
EXCL_LINES = None


def _setup_lines(exclude):
    global CODE_LINES, EXEC_LINES, EXCL_LINES
    if CODE_LINES is None or EXEC_LINES is None or EXCL_LINES is None:
        source = SRCDIR / "mpi4py" / "MPI.c"
        CODE_LINES, EXEC_LINES, EXCL_LINES = _parse_c_file(source, exclude)


def _parse_c_file(c_file, exclude_list):
    match_filetab_begin = "static const char *__pyx_f[] = {"
    match_filetab_begin = re.compile(re.escape(match_filetab_begin)).match
    match_filetab_entry = re.compile(r' *"(.*)",').match
    match_source_path_line = re.compile(r' */[*] +"(.*)":([0-9]+)$').match
    match_current_code_line = re.compile(r" *[*] (.*) # <<<<<<+$").match
    match_comment_end = re.compile(r" *[*]/$").match
    match_trace_line = re.compile(
        r" *__Pyx_TraceLine\((\d+),\d+,__PYX_ERR\((\d+),"
    ).match
    not_executable = re.compile(
        "|".join([
            r"\s*c(?:type)?def\s+"
            r"(?:(?:public|external)\s+)?"
            r"(?:struct|union|enum|class)"
            r"(\s+[^:]+|)\s*:",
        ])
    ).match
    if exclude_list:
        line_is_excluded = re.compile(
            "|".join([rf"(?:{regex})" for regex in exclude_list])
        ).search
    else:

        def line_is_excluded(_):
            return False

    filetab = []
    modinit = False
    code_lines = defaultdict(dict)
    exec_lines = defaultdict(dict)
    executable_lines = defaultdict(set)
    excluded_lines = defaultdict(set)

    with pathlib.Path(c_file).open(encoding="utf-8") as lines:
        lines = iter(lines)
        for line in lines:
            if match_filetab_begin(line):
                for line in lines:
                    match = match_filetab_entry(line)
                    if not match:
                        break
                    filename = match.group(1)
                    filetab.append(filename)
            match = match_source_path_line(line)
            if not match:
                if '__Pyx_TraceCall("__Pyx_PyMODINIT_FUNC ' in line:
                    modinit = True
                if "__Pyx_TraceLine(" in line:
                    trace_line = match_trace_line(line)
                    if trace_line:
                        lineno, fid = map(int, trace_line.groups())
                        executable_lines[filetab[fid]].add(lineno)
                continue
            filename, lineno = match.groups()
            lineno = int(lineno)
            for comment_line in lines:
                match = match_current_code_line(comment_line)
                if match:
                    code_line = match.group(1).rstrip()
                    if not_executable(code_line):
                        break
                    if line_is_excluded(code_line):
                        excluded_lines[filename].add(lineno)
                        break
                    code_lines[filename][lineno] = code_line
                    if modinit:
                        exec_lines[filename][lineno] = code_line
                    break
                if match_comment_end(comment_line):
                    # unexpected comment format - false positive?
                    break

    # Remove lines that generated code but are not traceable.

    for filename, lines in code_lines.items():
        dead_lines = set(lines).difference(executable_lines.get(filename, ()))
        for lineno in dead_lines:
            del lines[lineno]

    for filename, lines in exec_lines.items():
        dead_lines = set(lines).difference(executable_lines.get(filename, ()))
        for lineno in dead_lines:
            del lines[lineno]

    return code_lines, exec_lines, excluded_lines


def coverage_init(reg, options):  # noqa: ARG001
    plugin = CythonCoveragePlugin()
    reg.add_configurer(plugin)
    reg.add_file_tracer(plugin)



import os
import pathlib
import re
import shutil
import subprocess as sp
import sys

import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI

mod_file = pathlib.Path(MPI.__file__)
pxd_file = mod_file.parent / "libmpi.pxd"

mpi_small_count_allowed = [
    "MPI_Op_create",
]

mpi_fortran = [
    "MPI_Type_create_f90_integer",
    "MPI_Type_create_f90_real",
    "MPI_Type_create_f90_complex",
]

mpi_deprecated = [
    "MPI_Attr_delete",
    "MPI_Attr_get",
    "MPI_Attr_put",
    "MPI_Info_get",
    "MPI_Info_get_valuelen",
    "MPI_Keyval_create",
    "MPI_Keyval_free",
]

mpi_removed = [
    "MPI_Address",
    "MPI_Errhandler_create",
    "MPI_Errhandler_get",
    "MPI_Errhandler_set",
    "MPI_Type_extent",
    "MPI_Type_hindexed",
    "MPI_Type_hvector",
    "MPI_Type_lb",
    "MPI_Type_struct",
    "MPI_Type_ub",
]

mpi_missing = []

if MPI.Get_version() < (4, 1):
    mpi_missing += [
        f"MPI_Remove_error_{func}" for func in ("class", "code", "string")
    ]
    mpi_missing += [
        f"MPI_Status_{func}_{attr}"
        for func in ("get", "set")
        for attr in ("source", "tag", "error")
    ]
    mpi_missing += [
        f"MPI_Request_get_status_{func}" for func in ("any", "all", "some")
    ]
    mpi_missing += [f"MPI_Buffer_{func}" for func in ("flush", "iflush")] + [
        f"MPI_{cls}_{func}_buffer{kind}"
        for cls in ("Comm", "Session")
        for func in ("attach", "detach", "flush", "iflush")
        for kind in (("",) if "flush" in func else ("", "_c"))
    ]
    mpi_missing += [
        "MPI_Type_get_value_index",
        "MPI_Get_hw_resource_info",
    ]

if MPI.Get_version() < (5, 0):
    mpi_missing += [
        "MPI_Abi_get_version",
        "MPI_Abi_get_info",
        "MPI_Abi_get_fortran_info",
    ]
    for _type in (
        "Comm",
        "Errhandler",
        "File",
        "Group",
        "Info",
        "Message",
        "Op",
        "Request",
        "Session",
        "Type",
        "Win",
    ):
        mpi_missing += [
            f"MPI_{_type}_fromint",
            f"MPI_{_type}_toint",
        ]
else:
    mpi_missing += [
        "MPI_Status_c2f",
        "MPI_Status_f2c",
    ]
if MPI.Get_version() < (6, 0):
    mpi_missing += [
        "MPI_Comm_ack_failed",
        "MPI_Comm_agree",
        "MPI_Comm_get_failed",
        "MPI_Comm_iagree",
        "MPI_Comm_is_revoked",
        "MPI_Comm_ishrink",
        "MPI_Comm_revoke",
        "MPI_Comm_shrink",
    ]
name, version = MPI.get_vendor()
if name == "MPICH" and version < (4, 0, 3):
    mpi_missing += [
        "MPI_Status_set_elements_c",
    ]


@unittest.skipMPI("MPI(<4.0)")
class TestMPIAPI(unittest.TestCase):
    #
    MPINAME = r"MPI_[A-Z][a-z0-9_]+"

    def get_api_symbols(self):
        regex = re.compile(rf"^\s*int\s+({self.MPINAME})\s*\(")
        api_symbols = set()
        with pxd_file.open(encoding="utf-8") as fh:
            for line in fh:
                match = regex.search(line)
                if match:
                    sym = match.groups()[0]
                    api_symbols.add(sym)
        return api_symbols

    def get_mod_symbols(self):
        nm = shutil.which("nm")
        nm_flags = ["-Pu"]
        if sys.platform == "darwin":
            nm_flags.append("-m")
        if sys.platform == "linux":
            nm_flags.append("-D")
        cmd = [nm, *nm_flags, mod_file]
        out = sp.check_output(cmd, close_fds=False)
        nm_output = out.decode()

        mpiname = self.MPINAME
        if sys.platform == "darwin":
            regex = re.compile(
                rf"^\s+\(undefined\)\s+"
                r"(?:(?P<kind>w)eak)?\s*external\s+"
                rf"_(?P<name>{mpiname}).*$"
            )
        else:
            regex = re.compile(
                rf"^_?(?P<name>{mpiname})\s+"
                r"(?P<kind>[UVvWw]).*$"
            )
        mod_symbols = set()
        weak_symbols = set()
        for line in nm_output.split("\n"):
            match = regex.search(line)
            if match:
                sym = match.group("name")
                knd = match.group("kind") or "U"
                mod_symbols.add(sym)
                if knd.lower() in ("v", "w"):
                    weak_symbols.add(sym)
        for sym in weak_symbols:
            if sym.endswith("_c"):
                if sym[:-2] not in mpi_small_count_allowed:
                    mod_symbols.discard(sym[:-2])
        return mod_symbols

    @unittest.skipIf(shutil.which("nm") is None, "nm")
    @unittest.skipUnless(hasattr(os, "posix_spawn"), "os.posix_spawn")
    def testLargeCountSymbols(self):
        api_symbols = self.get_api_symbols()
        large_count = {sym for sym in api_symbols if sym.endswith("_c")}
        mpi_symbols = self.get_mod_symbols()

        for sym in mpi_small_count_allowed:
            self.assertIn(sym, mpi_symbols)
        mpi_symbols.difference_update(mpi_small_count_allowed)

        small_count = {sym[:-2] for sym in large_count}
        bad_symbols = set.intersection(small_count, mpi_symbols)
        self.assertFalse(bad_symbols)

    @unittest.skipIf(shutil.which("nm") is None, "nm")
    @unittest.skipUnless(hasattr(os, "posix_spawn"), "os.posix_spawn")
    def testSymbolCoverage(self):
        api_symbols = self.get_api_symbols()
        mod_symbols = self.get_mod_symbols()
        self.assertTrue(api_symbols)
        self.assertTrue(mod_symbols)

        uncovered = set.difference(api_symbols, mod_symbols)
        uncovered.difference_update(mpi_deprecated)
        uncovered.difference_update(mpi_removed)
        uncovered.difference_update(mpi_missing)
        for sym in mod_symbols:
            if sym.endswith("_c"):
                uncovered.discard(sym[:-2])
                uncovered.discard(sym[:-2] + "_x")
            if sym.endswith("_x"):
                uncovered.discard(sym[:-2])
                uncovered.discard(sym[:-2] + "_c")
        if not testutil.has_datatype(MPI.INTEGER):
            uncovered.difference_update(mpi_fortran)
        if name == "MPICH" and MPI.DISPLACEMENT_CURRENT == 0:
            mpiio = re.compile(r"MPI_(File_.*|Register_datarep)")
            uncovered = set(filter(lambda s: not mpiio.match(s), uncovered))
        self.assertFalse(uncovered)


if __name__ == "__main__":
    unittest.main()



from importlib import import_module
from os import environ
from warnings import catch_warnings

from mpi4py import rc

ATTRS = {a for a in dir(rc) if not a.startswith("_")}
VALUE = -123456789

for attr in ATTRS:
    environ.pop(f"MPI4PY_RC_{attr.upper()}", None)
    setattr(rc, attr, VALUE)

with catch_warnings(record=True) as warnings:
    import_module("mpi4py.MPI")

template = "mpi4py.rc.{}: unexpected value {}"
expected = sorted({template.format(attr, repr(VALUE)) for attr in ATTRS})
messages = sorted({str(entry.message) for entry in warnings})
assert messages == expected, "\n" + "\n".join(messages)



#!/usr/bin/env python
import os
import sys

import delocate.cmd.delocate_wheel as mod
from delocate.delocating import delocate_wheel, filter_system_libs

libmpi = []
for name in ("mpi", "pmpi", "open-pal", "open-rte"):
    for version in (1, 12, 20, 40):
        libmpi.append(f"lib{name}.{version}.dylib")
libmpi = set(libmpi)


def filter_libs(filename):
    if not filter_system_libs(filename):
        return False
    if os.path.basename(filename) in libmpi:
        return False
    return True


def delocate_wheel_patched(*args, **kwargs):
    kwargs["copy_filt_func"] = filter_libs
    return delocate_wheel(*args, **kwargs)


mod.delocate_wheel = delocate_wheel_patched

dyldname = "DYLD_FALLBACK_LIBRARY_PATH"
dyldpath = os.environ.get(dyldname, "").split(":")
dyldpath.append("/usr/local/lib")
dyldpath.append("/opt/homebrew/lib")
dyldpath.append("/opt/local/lib")
os.environ[dyldname] = ":".join(dyldpath)

if __name__ == "__main__":
    sys.exit(mod.main())



import contextlib
import importlib
import os
import pathlib
import shutil
import sys

# ---

BACKENDS = {
    "setuptools": "setuptools.build_meta",
    "skbuild": "scikit_build_core.setuptools.build_meta",
    "mesonpy": "mesonpy",
}


def get_build_backend_name(name=None):
    if name is None:
        name = os.environ.get("MPI4PY_BUILD_BACKEND", "")
        name = name.lower().replace("_", "-")
    if name in ("default", ""):
        for name, module_name in BACKENDS.items():
            if name != "setuptools":
                if module_name in sys.modules:
                    return name
        return "setuptools"
    if name in ("setuptools", "setup"):
        return "setuptools"
    if name in ("scikit-build-core", "scikit-build", "skbuild", "cmake"):
        return "skbuild"
    if name in ("meson-python", "mesonpy", "meson"):
        return "mesonpy"
    raise RuntimeError(f"Unknown build backend {name!r}")


def build_backend(name=None):
    name = get_build_backend_name(name)
    return importlib.import_module(BACKENDS[name])


def read_build_requires(name):
    confdir = pathlib.Path(__file__).resolve().parent
    basename = f"requirements-build-{name}.txt"
    filename = confdir / basename
    with filename.open(encoding="utf-8") as f:
        return [req for req in map(str.strip, f) if req]


def get_backend_requires_fast(backend, dist, config_settings=None):
    if backend is None or isinstance(backend, str):
        try:
            backend = build_backend(backend)
        except ImportError:
            return None
    requires = []
    get_requires = getattr(backend, f"get_requires_for_build_{dist}", None)
    if get_requires is not None:
        requires += get_requires(config_settings)
    return requires


def get_backend_requires_hook(name, dist, config_settings=None):
    try:
        from pyproject_hooks import BuildBackendHookCaller
    except ImportError:
        from pep517.wrappers import Pep517HookCaller as BuildBackendHookCaller
    try:
        from build.env import DefaultIsolatedEnv
    except ImportError:
        from build.env import IsolatedEnvBuilder

        class DefaultIsolatedEnv(IsolatedEnvBuilder):
            def __enter__(self):
                env = super().__enter__()
                env.python_executable = env.executable

                def make_extra_environ():
                    path = os.environ.get("PATH")
                    return {
                        "PATH": os.pathsep.join([env.scripts_dir, path])
                        if path is not None
                        else env.scripts_dir,
                    }

                env.make_extra_environ = make_extra_environ
                return env

    @contextlib.contextmanager
    def environment(path):
        environ_prev = [("PATH", os.environ["PATH"])]
        for prefix in ("_PYPROJECT_HOOKS", "PEP517"):
            for suffix in ("BUILD_BACKEND", "BACKEND_PATH"):
                key = f"{prefix}_{suffix}"
                if key in os.environ:
                    val = os.environ.pop(key)
                    environ_prev.append((key, val))
        os.environ["PATH"] = path
        try:
            yield None
        finally:
            os.environ.update(environ_prev)

    requires = read_build_requires(name)
    with DefaultIsolatedEnv() as env:
        path = env.make_extra_environ()["PATH"]
        python_executable = env.python_executable
        with environment(path):
            env.install(requires)
            hook = BuildBackendHookCaller(
                source_dir=pathlib.Path.cwd(),
                build_backend=BACKENDS[name],
                python_executable=python_executable,
            )
            requires += get_backend_requires_fast(hook, dist, config_settings)
    return requires


def get_requires_for_build(dist, config_settings=None):
    name = get_build_backend_name()
    requires = get_backend_requires_fast(name, dist, config_settings)
    if requires is None:
        requires = get_backend_requires_hook(name, dist, config_settings)
    if dist in ("wheel", "editable"):
        requires += read_build_requires("cython")
    return requires


# ---


def get_requires_for_build_sdist(config_settings=None):
    return get_requires_for_build("sdist", config_settings)


def build_sdist(
    sdist_directory,
    config_settings=None,
):
    return build_backend().build_sdist(
        sdist_directory,
        config_settings,
    )


# ---


def get_requires_for_build_wheel(config_settings=None):
    return get_requires_for_build("wheel", config_settings)


def prepare_metadata_for_build_wheel(
    metadata_directory,
    config_settings=None,
):
    return build_backend().prepare_metadata_for_build_wheel(
        metadata_directory,
        config_settings,
    )


def build_wheel(
    wheel_directory,
    config_settings=None,
    metadata_directory=None,
):
    return build_backend().build_wheel(
        wheel_directory,
        config_settings,
        metadata_directory,
    )


# ---


def get_requires_for_build_editable(config_settings=None):
    return get_requires_for_build("editable", config_settings)


def prepare_metadata_for_build_editable(
    metadata_directory,
    config_settings=None,
):
    return build_backend().prepare_metadata_for_build_editable(
        metadata_directory,
        config_settings,
    )


def build_editable(
    wheel_directory,
    config_settings=None,
    metadata_directory=None,
):
    return build_backend().build_editable(
        wheel_directory,
        config_settings,
        metadata_directory,
    )


# ---


def setup_env_mpicc():
    mpicc = shutil.which("mpicc")
    mpicc = os.environ.get("MPICC", mpicc)
    mpicc = os.environ.get("MPI4PY_BUILD_MPICC", mpicc)
    if not mpicc:
        return
    if " " in mpicc:
        mpicc = f'"{mpicc}"'
    if "CC" not in os.environ:
        os.environ["CC"] = mpicc


if get_build_backend_name() == "setuptools":
    try:
        import setuptools.build_meta as st_bm
    except ImportError:
        st_bm = None
    if not hasattr(st_bm, "get_requires_for_build_editable"):
        del get_requires_for_build_editable
    if not hasattr(st_bm, "prepare_metadata_for_build_editable"):
        del prepare_metadata_for_build_editable
    if not hasattr(st_bm, "build_editable"):
        del build_editable
    del st_bm

if get_build_backend_name() == "mesonpy":
    setup_env_mpicc()
    del prepare_metadata_for_build_wheel
    del prepare_metadata_for_build_editable

# ---



# -----------------------------------------------------------------------------

ctypedef fused mpi_scwf_t:
    MPI_Session
    MPI_Comm
    MPI_Win
    MPI_File

ctypedef fused mpi_ehfn_t:
    MPI_Session_errhandler_function
    MPI_Comm_errhandler_function
    MPI_Win_errhandler_function
    MPI_File_errhandler_function

cdef object errhdl_lock = Lock()
cdef list   errhdl_registry = [
    [None]*(1+32),  # Session
    [None]*(1+32),  # Comm
    [None]*(1+32),  # Win
    [None]*(1+32),  # File
]

cdef inline void errhdl_call_mpi(
    int index,
    mpi_scwf_t handle, int errcode,
) noexcept with gil:
    cdef object pyhandle = None
    cdef object registry = None
    # errors in user-defined error handler functions are unrecoverable
    try:
        if mpi_scwf_t is MPI_Session:
            pyhandle = fromhandle(handle)
            registry = errhdl_registry[0]
        if mpi_scwf_t is MPI_Comm:
            pyhandle = fromhandle(handle)
            registry = errhdl_registry[1]
        if mpi_scwf_t is MPI_Win:
            pyhandle = fromhandle(handle)
            registry = errhdl_registry[2]
        if mpi_scwf_t is MPI_File:
            pyhandle = fromhandle(handle)
            registry = errhdl_registry[3]
        try:
            registry[index](pyhandle, errcode)
        finally:
            if False: pass
            elif mpi_scwf_t is MPI_Session:
                (<Session>pyhandle).ob_mpi = MPI_SESSION_NULL
            elif mpi_scwf_t is MPI_Comm:
                (<Comm>pyhandle).ob_mpi = MPI_COMM_NULL
            elif mpi_scwf_t is MPI_Win:
                (<Win>pyhandle).ob_mpi = MPI_WIN_NULL
            elif mpi_scwf_t is MPI_File:
                (<File>pyhandle).ob_mpi = MPI_FILE_NULL
    except BaseException as exc:                                  # ~> uncovered
        PyErr_DisplayException(exc)                               # ~> uncovered
        PySys_WriteStderr(                                        # ~> uncovered
            b"Fatal Python error: %s\n",                          # ~> uncovered
            b"exception in user-defined error handler function",  # ~> uncovered
        )                                                         # ~> uncovered
        <void>MPI_Abort(MPI_COMM_WORLD, 1)                        # ~> uncovered


cdef inline void errhdl_call(
    int index,
    mpi_scwf_t handle, int errcode,
) noexcept nogil:
    # make it abort if Python has finalized
    if not Py_IsInitialized(): <void>MPI_Abort(MPI_COMM_WORLD, 1)
    # make it abort if module cleanup has been done
    if not py_module_alive():  <void>MPI_Abort(MPI_COMM_WORLD, 1)
    # make the actual GIL-safe Python call
    errhdl_call_mpi(index, handle, errcode)


@cython.callspec("MPIAPI")
cdef void errhdl_01(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  1, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_02(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  2, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_03(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  3, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_04(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  4, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_05(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  5, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_06(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  6, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_07(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  7, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_08(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  8, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_09(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call(  9, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_10(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 10, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_11(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 11, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_12(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 12, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_13(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 13, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_14(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 14, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_15(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 15, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_16(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 16, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_17(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 17, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_18(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 18, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_19(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 19, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_20(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 20, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_21(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 21, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_22(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 22, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_23(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 23, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_24(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 24, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_25(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 25, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_26(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 26, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_27(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 27, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_28(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 28, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_29(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 29, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_30(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 30, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_31(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 31, handle[0], errcode[0])


@cython.callspec("MPIAPI")
cdef void errhdl_32(mpi_scwf_t *handle, int *errcode, ...) noexcept nogil:
    errhdl_call( 32, handle[0], errcode[0])


cdef inline void errhdl_map(int index, mpi_ehfn_t **fn) noexcept nogil:
    if   index ==  1: fn[0] = errhdl_01
    elif index ==  2: fn[0] = errhdl_02
    elif index ==  3: fn[0] = errhdl_03
    elif index ==  4: fn[0] = errhdl_04
    elif index ==  5: fn[0] = errhdl_05
    elif index ==  6: fn[0] = errhdl_06
    elif index ==  7: fn[0] = errhdl_07
    elif index ==  8: fn[0] = errhdl_08
    elif index ==  9: fn[0] = errhdl_09
    elif index == 10: fn[0] = errhdl_10
    elif index == 11: fn[0] = errhdl_11
    elif index == 12: fn[0] = errhdl_12
    elif index == 13: fn[0] = errhdl_13
    elif index == 14: fn[0] = errhdl_14
    elif index == 15: fn[0] = errhdl_15
    elif index == 16: fn[0] = errhdl_16
    elif index == 17: fn[0] = errhdl_17
    elif index == 18: fn[0] = errhdl_18
    elif index == 19: fn[0] = errhdl_19
    elif index == 20: fn[0] = errhdl_20
    elif index == 21: fn[0] = errhdl_21
    elif index == 22: fn[0] = errhdl_22
    elif index == 23: fn[0] = errhdl_23
    elif index == 24: fn[0] = errhdl_24
    elif index == 25: fn[0] = errhdl_25
    elif index == 26: fn[0] = errhdl_26
    elif index == 27: fn[0] = errhdl_27
    elif index == 28: fn[0] = errhdl_28
    elif index == 29: fn[0] = errhdl_29
    elif index == 30: fn[0] = errhdl_30
    elif index == 31: fn[0] = errhdl_31
    elif index == 32: fn[0] = errhdl_32


cdef inline int errhdl_new(
    object function,
    mpi_ehfn_t **fn,
) except -1:
    # check whether the function is callable
    function.__call__
    # find a free slot in the registry
    # and register the Python function
    cdef list registry = None
    if mpi_ehfn_t is MPI_Session_errhandler_function:
        registry = errhdl_registry[0]
    if mpi_ehfn_t is MPI_Comm_errhandler_function:
        registry = errhdl_registry[1]
    if mpi_ehfn_t is MPI_Win_errhandler_function:
        registry = errhdl_registry[2]
    if mpi_ehfn_t is MPI_File_errhandler_function:
        registry = errhdl_registry[3]
    cdef int index = 0
    try:
        with errhdl_lock:
            index = registry.index(None, 1)
            registry[index] = function
    except ValueError:
        raise RuntimeError(
            "cannot create too many user-defined error handlers",
        )
    # map slot index to the associated C callback,
    # and return the slot index in the registry
    errhdl_map(index, fn)
    return index


@cython.linetrace(False)  # ~> TODO
cdef inline int errhdl_del(
    int *indexp,
    mpi_ehfn_t *fn,
) except -1:
    <void> fn  # unused
    # clear index value
    cdef int index = indexp[0]
    indexp[0] = 0
    # free slot in the registry
    cdef object registry = None
    if mpi_ehfn_t is MPI_Session_errhandler_function:
        registry = errhdl_registry[0]
    if mpi_ehfn_t is MPI_Comm_errhandler_function:
        registry = errhdl_registry[1]
    if mpi_ehfn_t is MPI_Win_errhandler_function:
        registry = errhdl_registry[2]
    if mpi_ehfn_t is MPI_File_errhandler_function:
        registry = errhdl_registry[3]
    with errhdl_lock:
        registry[index] = None
    return 0

# -----------------------------------------------------------------------------

cdef inline MPI_Errhandler options_get_errhandler() noexcept nogil:
    if options.errors == 0: return MPI_ERRHANDLER_NULL
    if options.errors == 1: return MPI_ERRORS_RETURN
    if options.errors == 2: return MPI_ERRORS_ABORT
    if options.errors == 3: return MPI_ERRORS_ARE_FATAL
    return MPI_ERRHANDLER_NULL  # ~> unreachable


cdef inline int options_set_errhandler(mpi_scwf_t handle) except -1 nogil:
    if handle == mpinull(handle): return 0
    cdef MPI_Errhandler errhandler = options_get_errhandler()
    if errhandler == MPI_ERRHANDLER_NULL: return 0
    if mpi_scwf_t is MPI_Session:
        CHKERR( MPI_Session_set_errhandler(handle, errhandler) )
    if mpi_scwf_t is MPI_Comm:
        CHKERR( MPI_Comm_set_errhandler(handle, errhandler) )
    if mpi_scwf_t is MPI_Win:
        CHKERR( MPI_Win_set_errhandler(handle, errhandler) )
    if mpi_scwf_t is MPI_File:
        CHKERR( MPI_File_set_errhandler(handle, errhandler) )
    return 0

# -----------------------------------------------------------------------------



import sys

import mpi4py

mpi4py.rc.initialize = False

from mpi4py import MPI  # noqa: E402


def test_error():
    for errcls in [
        getattr(MPI, attr) for attr in dir(MPI) if attr.startswith("ERR_")
    ]:
        if errcls == MPI.ERR_LASTCODE:
            continue

        errcod = MPI.Get_error_class(errcls)
        assert errcod > MPI.SUCCESS
        assert errcod == errcls

        errstr = MPI.Get_error_string(errcls)
        assert errstr

        if MPI.Get_version() >= (4, 0):
            excobj = MPI.Exception(errcls)
            excstr = str(excobj)
            assert excstr == errstr


def test_info():
    info = MPI.Info.Create()

    info.Dup().Free()

    info.Set("hello", "world")

    info.Dup().Free()

    key = info.Get_nthkey(0)
    assert key == "hello"

    value = info.Get("hello")
    assert value == "world"

    info.Delete("hello")

    nkeys = info.Get_nkeys()
    assert nkeys == 0

    value = info.Get("hello")
    assert value is None

    finfo = MPI.Info.f2py(info.py2f())
    assert info == finfo

    info.Free()


def test_session():
    for _ in range(2):
        session = MPI.Session.Init()

        for n in range(session.Get_num_psets()):
            pset_name = session.Get_nth_pset(n)
            info = session.Get_pset_info(pset_name)
            mpi_size = int(info.Get("mpi_size"))
            info.Free()
            assert mpi_size > 0

            group = session.Create_group(pset_name)
            assert mpi_size == group.Get_size()
            comm = MPI.Intracomm.Create_from_group(group)
            assert group.Get_size() == comm.Get_size()
            assert group.Get_rank() == comm.Get_rank()
            comm.Barrier()
            group.Free()
            comm.Free()

        session.Finalize()


def test_all():
    test_error()
    test_info()
    test_session()


def skip():
    name, version = MPI.get_vendor()
    if name == "MPICH":
        if version >= (5, 0, 0):
            return False
    if name == "Open MPI":
        if version >= (5, 1, 0):
            return False
    return True


if skip():
    sys.exit(0)

assert not MPI.Is_initialized()
assert not MPI.Is_finalized()

test_all()

assert not MPI.Is_initialized()
assert not MPI.Is_finalized()

MPI.Init()

assert MPI.Is_initialized()
assert not MPI.Is_finalized()

test_all()

assert MPI.Is_initialized()
assert not MPI.Is_finalized()

MPI.Finalize()

assert MPI.Is_initialized()
assert MPI.Is_finalized()

test_all()

assert MPI.Is_initialized()
assert MPI.Is_finalized()



#!/usr/bin/env python

if False:
    import mpi4py

    name = "name"  # lib{name}.so
    path = []
    mpi4py.profile(name, path=path)

from mpi4py import MPI

comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

src = rank - 1
dest = rank + 1
if rank == 0:
    src = size - 1
if rank == size - 1:
    dest = 0

try:
    from numpy import zeros

    a1 = zeros(1000000, "d")
    a2 = zeros(1000000, "d")
except ImportError:
    from array import array

    a1 = array("d", [0] * 1000)
    a1 *= 1000
    a2 = array("d", [0] * 1000)
    a2 *= 1000

comm.Sendrecv(
    sendbuf=a1,
    recvbuf=a2,
    source=src,
    dest=dest,
)

MPI.Request.Waitall([
    comm.Isend(a1, dest=dest),
    comm.Irecv(a2, source=src),
])



from libmpi import ffi, lib


def ring(comm, count=1, loop=1, skip=0):
    size_p = ffi.new("int*")
    rank_p = ffi.new("int*")
    lib.MPI_Comm_size(comm, size_p)
    lib.MPI_Comm_rank(comm, rank_p)
    size = size_p[0]
    rank = rank_p[0]

    source = (rank - 1) % size
    dest = (rank + 1) % size
    sbuf = ffi.new("unsigned char[]", [42] * count)
    rbuf = ffi.new("unsigned char[]", [0] * count)

    iterations = list(range(loop + skip))

    if size == 1:
        for i in iterations:
            if i == skip:
                tic = lib.MPI_Wtime()
            lib.MPI_Sendrecv(
                sbuf,
                count,
                lib.MPI_BYTE,
                dest,
                0,
                rbuf,
                count,
                lib.MPI_BYTE,
                source,
                0,
                comm,
                lib.MPI_STATUS_IGNORE,
            )
    else:
        if rank == 0:
            for i in iterations:
                if i == skip:
                    tic = lib.MPI_Wtime()
                lib.MPI_Send(sbuf, count, lib.MPI_BYTE, dest, 0, comm)
                lib.MPI_Recv(
                    rbuf,
                    count,
                    lib.MPI_BYTE,
                    source,
                    0,
                    comm,
                    lib.MPI_STATUS_IGNORE,
                )
        else:
            sbuf = rbuf
            for i in iterations:
                if i == skip:
                    tic = lib.MPI_Wtime()
                lib.MPI_Recv(
                    rbuf,
                    count,
                    lib.MPI_BYTE,
                    source,
                    0,
                    comm,
                    lib.MPI_STATUS_IGNORE,
                )
                lib.MPI_Send(sbuf, count, lib.MPI_BYTE, dest, 0, comm)
    toc = lib.MPI_Wtime()
    if rank == 0 and ffi.string(sbuf) != ffi.string(rbuf):
        import traceback
        import warnings

        try:
            warnings.warn("received message does not match!", stacklevel=2)
        except UserWarning:
            traceback.print_exc()
            lib.MPI_Abort(comm, 2)
    return toc - tic


def ringtest(comm):
    size = 1
    loop = 1
    skip = 0

    lib.MPI_Barrier(comm)
    elapsed = ring(comm, size, loop, skip)

    size_p = ffi.new("int*")
    rank_p = ffi.new("int*")
    lib.MPI_Comm_size(comm, size_p)
    lib.MPI_Comm_rank(comm, rank_p)
    comm_size = size_p[0]
    comm_rank = rank_p[0]

    if comm_rank == 0:
        print(
            f"time for {loop} loops = {elapsed:g} seconds "
            f"({comm_size:d} processes, {size:d} bytes)"
        )


def main():
    lib.MPI_Init(ffi.NULL, ffi.NULL)
    ringtest(lib.MPI_COMM_WORLD)
    lib.MPI_Finalize()


if __name__ == "__main__":
    main()



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com
"""Support for Future chaining."""
# pylint: disable=broad-exception-caught

# This implementation is heavily inspired in code written by
# Daniel Dotsenko [@dvdotsenko] [dotsa at hotmail.com]
# https://github.com/dvdotsenko/python-future-then

import functools
import threading
import weakref

from ._base import Future


class ThenableFuture(Future):
    """*Thenable* `Future` subclass."""

    def then(self, on_success=None, on_failure=None):
        """Return ``then(self, on_success, on_failure)``."""
        return then(self, on_success, on_failure)

    def catch(self, on_failure=None):
        """Return ``catch(self, on_failure)``."""
        return catch(self, on_failure)


def then(future, on_success=None, on_failure=None):
    """JavaScript-like (`Promises/A+`_) support for Future chaining.

    Args:
        future: Input future instance.
        on_success: Function to be called once the input future completes
            with success. Once the input future finish running with success,
            its result value is the input argument for *on_success*.
            If *on_success* returns a `Future` instance, the result future is
            chained to the output future.  Otherwise, the result of
            *on_success* is set as the result of the output future.
            If *on_success* is ``None``, the output future is resolved directly
            with the result of the input future.
        on_failure: Function to be called once the input future completes with
            failure. Once the input future finish running with failure, its
            exception value is the input argument for *on_failure*.
            If *on_failure* returns a `Future` instance, the result future is
            chained to the output future. Otherwise, if *on_failure* returns an
            `Exception` instance, it is set as the exception of the output
            future.  Otherwise, the result of *on_failure* is set as the result
            of the output future.
            If *on_failure* is ``None``, the output future is set as failed
            with the exception from the input future.

    Returns:
        Output future instance to be completed once the input future is
        completed and either *on_success* or *on_failure* completes.

    .. _Promises/A+: https://promisesaplus.com/

    """
    return _chain(future, on_success, on_failure)


def catch(future, on_failure=None):
    """Close equivalent to ``then(future, None, on_failure)``.

    Args:
        future: Input future instance.
        on_failure: Function to be called once the input future completes with
            failure. Once the input future finish running with failure, its
            exception value is the input argument for *on_failure*.
            If *on_failure* returns a `Future` instance, the result future is
            chained to the output future. Otherwise, if *on_failure* returns an
            `Exception` instance, it is set as the exception of the output
            future.  Otherwise, the result of *on_failure* is set as the result
            of the output future.
            If *on_failure* is ``None``, the output future is set as failed
            with the exception from the input future.

    Returns:
        Output future instance to be completed once the input future is
        completed and *on_failure* completes.

    """
    if on_failure is None:
        return then(future, None, lambda _: None)
    return then(future, None, on_failure)


def _chain(future, on_success=None, on_failure=None):
    new_future = future.__class__()
    done_cb = functools.partial(
        _chain_resolve,
        [new_future],
        on_success=on_success,
        on_failure=on_failure,
    )
    future.add_done_callback(done_cb)
    return new_future


_chain_log_lock = threading.Lock()
_chain_log_registry = weakref.WeakKeyDictionary()


def _chain_check_cycle(new_future, future):
    if new_future is future:
        raise RuntimeError(
            f"chain cycle detected: Future {future} chained with itself"
        )
    with _chain_log_lock:
        registry = _chain_log_registry
        try:
            log = registry[new_future]
        except KeyError:
            log = weakref.WeakSet()
            registry[new_future] = log
        if future in log:
            raise RuntimeError(
                f"chain cycle detected: Future {future} already in chain"
            )
        log.add(future)


def _chain_resolve_future(new_future, old_future):
    _chain_check_cycle(new_future, old_future)
    done_cb = functools.partial(_chain_resolve, [new_future])
    old_future.add_done_callback(done_cb)


def _chain_resolve_success(new_future, result, on_success=None):
    try:
        if on_success:
            result = on_success(result)
        if isinstance(result, Future):
            _chain_resolve_future(new_future, result)
        else:
            new_future.set_result(result)
    finally:
        del new_future, result, on_success


def _chain_resolve_failure(new_future, reason, on_failure=None):
    try:
        if on_failure:
            try:
                reason = on_failure(reason)
                if isinstance(reason, Future):
                    _chain_resolve_future(new_future, reason)
                elif isinstance(reason, BaseException):
                    new_future.set_exception(reason)
                else:
                    new_future.set_result(reason)
            except BaseException as exception:
                new_future.set_exception(exception)
        else:
            new_future.set_exception(reason)
    finally:
        del new_future, reason, on_failure


def _chain_resolve(ctx, old_future, on_success=None, on_failure=None):
    new_future = ctx.pop()
    if old_future.cancelled() or new_future.cancelled():
        new_future.cancel()
        new_future.set_running_or_notify_cancel()
        return
    try:
        if old_future.exception() is not None:
            _chain_resolve_failure(
                new_future,
                old_future.exception(),
                on_failure,
            )
        else:
            try:
                _chain_resolve_success(
                    new_future,
                    old_future.result(),
                    on_success,
                )
            except BaseException as exception:
                _chain_resolve_failure(
                    new_future,
                    exception,
                    on_failure,
                )
    finally:
        del new_future, old_future



import os
import pathlib
import platform
import tempfile

import arrayimpl
import mpiunittest as unittest

from mpi4py import MPI


def maketemp(prefix):
    fd, name = tempfile.mkstemp(prefix=prefix)
    os.close(fd)
    return name


def arrayimpl_loop_io():
    impi = unittest.mpi_predicate("impi(>=2021.12.0)")
    openmpi = unittest.mpi_predicate("openmpi(<4.2.0)")
    is_i386 = platform.machine() in ("i386", "i686")
    is_win = os.name == "nt"
    for array, typecode in arrayimpl.loop():
        if unittest.is_mpi_gpu("mvapich", array):
            continue
        if openmpi and is_i386 and typecode in ("g", "G"):
            continue
        if impi and is_win and typecode in ("l", "L", "g"):
            continue
        yield array, typecode


scalar = arrayimpl.scalar


class BaseTestIO:
    #
    COMM = MPI.COMM_NULL
    FILE = MPI.FILE_NULL

    prefix = "mpi4py-"

    def setUp(self):
        comm = self.COMM
        world_size = MPI.COMM_WORLD.Get_size()
        world_rank = MPI.COMM_WORLD.Get_rank()
        prefix = self.prefix
        if comm.Get_size() < world_size:
            prefix += f"{world_rank}-"
        fname = None
        if comm.Get_rank() == 0:
            fname = maketemp(prefix)
            fname = pathlib.Path(fname)
        fname = comm.bcast(fname, 0)
        amode = MPI.MODE_RDWR | MPI.MODE_CREATE
        amode |= MPI.MODE_DELETE_ON_CLOSE
        amode |= MPI.MODE_UNIQUE_OPEN
        info = MPI.INFO_NULL
        try:
            self.FILE = MPI.File.Open(comm, fname, amode, info)
        except Exception:
            if comm.Get_rank() == 0:
                fname.unlink()
            raise

    def tearDown(self):
        if self.FILE:
            self.FILE.Close()
        self.COMM.Barrier()


class BaseTestIOBasic(BaseTestIO):
    #
    # non-collective

    def testReadWriteAt(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Write_at(count * rank, wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Read_at(count * rank, rbuf.as_mpi_c(count))
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testIReadIWriteAt(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Iwrite_at(count * rank, wbuf.as_raw()).Wait()
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Iread_at(count * rank, rbuf.as_mpi_c(count)).Wait()
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testReadWrite(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                for r in range(size):
                    if r == rank:
                        fh.Seek(0, MPI.SEEK_SET)
                        fh.Write(wbuf.as_raw())
                    fh.Sync()
                    comm.Barrier()
                    fh.Sync()
                    for n in range(len(wbuf)):
                        rbuf = array(-1, typecode, n + 1)
                        fh.Seek(0, MPI.SEEK_SET)
                        fh.Read(rbuf.as_mpi_c(n))
                        for value in rbuf[:-1]:
                            self.assertEqual(value, scalar(42))
                        self.assertEqual(rbuf[-1], scalar(-1))
                    comm.Barrier()

    def testIReadIWrite(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                for r in range(size):
                    if r == rank:
                        fh.Seek(0, MPI.SEEK_SET)
                        fh.Iwrite(wbuf.as_raw()).Wait()
                    fh.Sync()
                    comm.Barrier()
                    fh.Sync()
                    for n in range(len(wbuf)):
                        rbuf = array(-1, typecode, n + 1)
                        fh.Seek(0, MPI.SEEK_SET)
                        fh.Iread(rbuf.as_mpi_c(n)).Wait()
                        for value in rbuf[:-1]:
                            self.assertEqual(value, scalar(42))
                        self.assertEqual(rbuf[-1], scalar(-1))
                    comm.Barrier()

    def testReadWriteShared(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(rank % 42, typecode, count)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Write_shared(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Read_shared(rbuf.as_mpi_c(count))
                for value in rbuf[:-1]:
                    self.assertTrue(0 <= value < 42)
                    self.assertEqual(value, rbuf[0])
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testIReadIWriteShared(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(rank % 42, typecode, count)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Iwrite_shared(wbuf.as_raw()).Wait()
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Iread_shared(rbuf.as_mpi_c(count)).Wait()
                for value in rbuf[:-1]:
                    self.assertTrue(0 <= value < 42)
                    self.assertEqual(value, rbuf[0])
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    # collective

    def testReadWriteAtAll(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Write_at_all(count * rank, wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Read_at_all(count * rank, rbuf.as_mpi_c(count))
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testIReadIWriteAtAll(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                try:  # MPI 3.1
                    etype = array.TypeMap[typecode]
                    fh.Set_size(0)
                    fh.Set_view(0, etype)
                    count = 13
                    wbuf = array(42, typecode, count)
                    fh.Iwrite_at_all(count * rank, wbuf.as_raw()).Wait()
                    fh.Sync()
                    comm.Barrier()
                    fh.Sync()
                    rbuf = array(-1, typecode, count + 1)
                    fh.Iread_at_all(count * rank, rbuf.as_mpi_c(count)).Wait()
                    for value in rbuf[:-1]:
                        self.assertEqual(value, scalar(42))
                    self.assertEqual(rbuf[-1], scalar(-1))
                    comm.Barrier()
                except NotImplementedError:
                    if MPI.Get_version() >= (3, 1):
                        raise
                    self.skipTest("mpi-iwrite_at_all")

    def testReadWriteAtAllBeginEnd(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Write_at_all_begin(count * rank, wbuf.as_raw())
                fh.Write_at_all_end(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Read_at_all_begin(count * rank, rbuf.as_mpi_c(count))
                fh.Read_at_all_end(rbuf.as_raw())
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testReadWriteAll(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Seek(count * rank, MPI.SEEK_SET)
                fh.Write_all(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Seek(count * rank, MPI.SEEK_SET)
                fh.Read_all(rbuf.as_mpi_c(count))
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testIReadIWriteAll(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                try:  # MPI 3.1
                    etype = array.TypeMap[typecode]
                    fh.Set_size(0)
                    fh.Set_view(0, etype)
                    count = 13
                    wbuf = array(42, typecode, count)
                    fh.Seek(count * rank, MPI.SEEK_SET)
                    fh.Iwrite_all(wbuf.as_raw()).Wait()
                    fh.Sync()
                    comm.Barrier()
                    fh.Sync()
                    rbuf = array(-1, typecode, count + 1)
                    fh.Seek(count * rank, MPI.SEEK_SET)
                    fh.Iread_all(rbuf.as_mpi_c(count)).Wait()
                    for value in rbuf[:-1]:
                        self.assertEqual(value, scalar(42))
                    self.assertEqual(rbuf[-1], scalar(-1))
                    comm.Barrier()
                except NotImplementedError:
                    if MPI.Get_version() >= (3, 1):
                        raise
                    self.skipTest("mpi-iwrite_all")

    def testReadWriteAllBeginEnd(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(42, typecode, count)
                fh.Seek(count * rank, MPI.SEEK_SET)
                fh.Write_all_begin(wbuf.as_raw())
                fh.Write_all_end(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Seek(count * rank, MPI.SEEK_SET)
                fh.Read_all_begin(rbuf.as_mpi_c(count))
                fh.Read_all_end(rbuf.as_raw())
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testReadWriteOrdered(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(rank % 42, typecode, count)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Write_ordered(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Read_ordered(rbuf.as_mpi_c(count))
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(rank % 42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()

    def testReadWriteOrderedBeginEnd(self):
        comm = self.COMM
        comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                etype = array.TypeMap[typecode]
                fh.Set_size(0)
                fh.Set_view(0, etype)
                count = 13
                wbuf = array(rank % 42, typecode, count)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Write_ordered_begin(wbuf.as_raw())
                fh.Write_ordered_end(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Sync()
                rbuf = array(-1, typecode, count + 1)
                fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Read_ordered_begin(rbuf.as_mpi_c(count))
                fh.Read_ordered_end(rbuf.as_raw())
                for value in rbuf[:-1]:
                    self.assertEqual(value, scalar(rank % 42))
                self.assertEqual(rbuf[-1], scalar(-1))
                comm.Barrier()


class BaseTestIOView(BaseTestIO):
    #
    def _test_contiguous(self, combiner):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                comm.Barrier()
                fh.Set_size(0)
                btype = array.TypeMap[typecode]
                if combiner == MPI.COMBINER_NAMED:
                    etype = btype
                    ftype = btype
                if combiner == MPI.COMBINER_DUP:
                    etype = btype.Dup().Commit()
                    ftype = etype.Dup().Commit()
                if combiner == MPI.COMBINER_CONTIGUOUS:
                    etype = btype
                    ftype = etype.Create_contiguous(7).Commit()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    wval = 10 * (rank + 1) + i
                    wbuf = array(wval, typecode, 7)
                    fh.Write_ordered(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    rval = 10 * (rank + 1) + i
                    rbuf = array(0, typecode, 7)
                    fh.Read_ordered(rbuf.as_raw())
                    for value in rbuf:
                        self.assertEqual(value, scalar(rval))
                if ftype != btype:
                    ftype.Free()
                fh.Set_view(0, etype, etype)
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                for i in range(3):
                    for r in range(size):
                        rval = 10 * (r + 1) + i
                        rbuf = array(0, typecode, 7)
                        fh.Read_all(rbuf.as_raw())
                        for value in rbuf:
                            self.assertEqual(value, scalar(rval))
                if etype != btype:
                    etype.Free()

    def _test_strided(self, combiner):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                comm.Barrier()
                fh.Set_size(0)
                etype = array.TypeMap[typecode]
                esize = etype.size
                index1 = [0, 2, 4, 6]
                index2 = [1, 3, 5]
                if combiner in (
                    MPI.COMBINER_VECTOR,
                    MPI.COMBINER_HVECTOR,
                ):
                    if combiner == MPI.COMBINER_VECTOR:
                        ftype1 = etype.Create_vector(4, 1, 2).Commit()
                    if combiner == MPI.COMBINER_HVECTOR:
                        ftype1 = etype.Create_hvector(4, 1, esize * 2).Commit()
                    fbase2 = etype.Create_indexed([1] * 3, index2)
                    ftype2 = fbase2.Create_resized(0, ftype1.extent).Commit()
                    fbase2.Free()
                if combiner in (
                    MPI.COMBINER_INDEXED,
                    MPI.COMBINER_INDEXED_BLOCK,
                    MPI.COMBINER_HINDEXED,
                    MPI.COMBINER_HINDEXED_BLOCK,
                ):
                    INDEXED = MPI.COMBINER_INDEXED
                    INDEXED_BLOCK = MPI.COMBINER_INDEXED_BLOCK
                    HINDEXED = MPI.COMBINER_HINDEXED
                    HINDEXED_BLOCK = MPI.COMBINER_HINDEXED_BLOCK
                    if combiner == INDEXED:
                        Create = MPI.Datatype.Create_indexed
                    if combiner == HINDEXED:
                        Create = MPI.Datatype.Create_hindexed
                    if combiner == INDEXED_BLOCK:
                        Create = MPI.Datatype.Create_indexed_block
                    if combiner == HINDEXED_BLOCK:
                        Create = MPI.Datatype.Create_hindexed_block
                    if combiner in (INDEXED, HINDEXED):
                        blens1 = [1] * 4
                        blens2 = [1] * 3
                    if combiner in (INDEXED_BLOCK, HINDEXED_BLOCK):
                        blens1 = 1
                        blens2 = 1
                    if combiner in (INDEXED, INDEXED_BLOCK):
                        disps1 = index1
                        disps2 = index2
                    if combiner in (HINDEXED, HINDEXED_BLOCK):
                        disps1 = [esize * i for i in index1]
                        disps2 = [esize * i for i in index2]
                    ftype1 = Create(etype, blens1, disps1).Commit()
                    fbase2 = Create(etype, blens2, disps2)
                    ftype2 = fbase2.Create_resized(0, ftype1.extent).Commit()
                    fbase2.Free()
                if combiner == MPI.COMBINER_STRUCT:
                    ftype1 = MPI.Datatype.Create_struct(
                        [1] * 4,
                        [esize * i for i in index1],
                        [etype] * 4,
                    ).Commit()
                    fbase2 = MPI.Datatype.Create_struct(
                        [1] * 3,
                        [esize * i for i in index2],
                        [etype] * 3,
                    )
                    ftype2 = fbase2.Create_resized(
                        0,
                        ftype1.extent,
                    ).Commit()
                    fbase2.Free()
                #
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Set_view(0, etype, ftype1)
                for i in range(3):
                    wval = 10 * (rank + 1) + i
                    warg = [wval + j for j in range(0, 7, 2)]
                    wbuf = array(warg, typecode)
                    fh.Write_ordered(wbuf.as_raw())
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Set_view(0, etype, ftype2)
                for i in range(3):
                    wval = 10 * (rank + 1) + i
                    warg = [wval + j for j in range(1, 7, 2)]
                    wbuf = array(warg, typecode)
                    fh.Write_ordered(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Set_view(0, etype, ftype1)
                for i in range(3):
                    rval = 10 * (rank + 1) + i
                    rbuf = array(0, typecode, 4)
                    fh.Read_ordered(rbuf.as_raw())
                    for value, j in zip(rbuf, range(0, 7, 2)):
                        self.assertEqual(value, scalar(rval + j))
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Set_view(0, etype, ftype2)
                for i in range(3):
                    rval = 10 * (rank + 1) + i
                    rbuf = array(0, typecode, 3)
                    fh.Read_ordered(rbuf.as_raw())
                    for value, j in zip(rbuf, range(1, 7, 2)):
                        self.assertEqual(value, scalar(rval + j))
                ftype1.Free()
                ftype2.Free()
                if unittest.is_mpi("openmpi(<4.0.0)"):
                    fh.Seek_shared(0, MPI.SEEK_SET)
                fh.Set_view(0, etype, etype)
                for i in range(3):
                    for r in range(size):
                        rval = 10 * (r + 1) + i
                        rbuf = array(0, typecode, 7)
                        fh.Read_all(rbuf.as_raw())
                        for j, value in enumerate(rbuf):
                            self.assertEqual(value, scalar(rval + j))

    def testNamed(self):
        self._test_contiguous(MPI.COMBINER_NAMED)

    def testDup(self):
        self._test_contiguous(MPI.COMBINER_DUP)

    def testContiguous(self):
        self._test_contiguous(MPI.COMBINER_CONTIGUOUS)

    def testVector(self):
        self._test_strided(MPI.COMBINER_VECTOR)

    def testHVector(self):
        self._test_strided(MPI.COMBINER_HVECTOR)

    def testIndexed(self):
        self._test_strided(MPI.COMBINER_INDEXED)

    def testHIndexed(self):
        self._test_strided(MPI.COMBINER_HINDEXED)

    def testIndexedBlock(self):
        self._test_strided(MPI.COMBINER_INDEXED_BLOCK)

    def testHIndexedBlock(self):
        self._test_strided(MPI.COMBINER_HINDEXED_BLOCK)

    def testStruct(self):
        self._test_strided(MPI.COMBINER_STRUCT)

    def testSubarray(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                comm.Barrier()
                fh.Set_size(0)
                etype = array.TypeMap[typecode]
                ftype = etype.Create_subarray(
                    [size * 7, 5],
                    [7, 5],
                    [rank * 7, 0],
                ).Commit()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    wval = 10 * (rank + 1) + i
                    wbuf = array(wval, typecode, 7 * 5)
                    fh.Write_all(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    rval = 10 * (rank + 1) + i
                    rbuf = array(0, typecode, 7 * 5)
                    fh.Read_all(rbuf.as_raw())
                    for value in rbuf:
                        self.assertEqual(value, scalar(rval))
                ftype.Free()
                fh.Set_view(0, etype, etype)
                for i in range(3):
                    for r in range(size):
                        rval = 10 * (r + 1) + i
                        rbuf = array(0, typecode, 7 * 5)
                        fh.Read_all(rbuf.as_raw())
                        for value in rbuf:
                            self.assertEqual(value, scalar(rval))

    def testDarrayBlock(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        block = MPI.DISTRIBUTE_BLOCK
        none = MPI.DISTRIBUTE_NONE
        dflt = MPI.DISTRIBUTE_DFLT_DARG
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                comm.Barrier()
                fh.Set_size(0)
                etype = array.TypeMap[typecode]
                ftype = etype.Create_darray(
                    size,
                    rank,
                    [size * 7, 5],
                    [block, none],
                    [dflt, dflt],
                    [size, 1],
                ).Commit()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    wval = 10 * (rank + 1) + i
                    wbuf = array(wval, typecode, 7 * 5)
                    fh.Write_all(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    rval = 10 * (rank + 1) + i
                    rbuf = array(0, typecode, 7 * 5)
                    fh.Read_all(rbuf.as_raw())
                    for value in rbuf:
                        self.assertEqual(value, scalar(rval))
                ftype.Free()
                fh.Set_view(0, etype, etype)
                for i in range(3):
                    for r in range(size):
                        for _j in range(7):
                            rval = 10 * (r + 1) + i
                            rbuf = array(0, typecode, 5)
                            fh.Read_all(rbuf.as_raw())
                            for value in rbuf:
                                self.assertEqual(value, scalar(rval))

    def testDarrayCyclic(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        fh = self.FILE
        cyclic = MPI.DISTRIBUTE_CYCLIC
        none = MPI.DISTRIBUTE_NONE
        dflt = MPI.DISTRIBUTE_DFLT_DARG
        for array, typecode in arrayimpl_loop_io():
            with arrayimpl.test(self):
                comm.Barrier()
                fh.Set_size(0)
                etype = array.TypeMap[typecode]
                ftype = etype.Create_darray(
                    size,
                    rank,
                    [size * 7, 5],
                    [cyclic, none],
                    [1, dflt],
                    [size, 1],
                ).Commit()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    wval = 10 * (rank + 1) + i
                    wbuf = array(wval, typecode, 7 * 5)
                    fh.Write_all(wbuf.as_raw())
                fh.Sync()
                comm.Barrier()
                fh.Set_view(0, etype, ftype)
                for i in range(3):
                    rval = 10 * (rank + 1) + i
                    rbuf = array(0, typecode, 7 * 5)
                    fh.Read_all(rbuf.as_raw())
                    for value in rbuf:
                        self.assertEqual(value, scalar(rval))
                ftype.Free()
                fh.Set_view(0, etype, etype)
                for i in range(3):
                    for _j in range(7):
                        for r in range(size):
                            rval = 10 * (r + 1) + i
                            rbuf = array(0, typecode, 5)
                            fh.Read_all(rbuf.as_raw())
                            for value in rbuf:
                                self.assertEqual(value, scalar(rval))


@unittest.skipMPI("MPICH1")
class TestIOBasicSelf(BaseTestIOBasic, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("openmpi(<2.2.0)")
@unittest.skipMPI("msmpi")
@unittest.skipMPI("MPICH2")
@unittest.skipMPI("MPICH1")
class TestIOBasicWorld(BaseTestIOBasic, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


@unittest.skipMPI("mpich(>=4.0.0,<4.1.0)")
@unittest.skipMPI("openmpi(<2.2.0)")
@unittest.skipMPI("MPICH1")
class TestIOViewSelf(BaseTestIOView, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("mpich(>=4.0.0,<4.1.0)")
@unittest.skipMPI("openmpi(<2.2.0)")
@unittest.skipMPI("msmpi")
@unittest.skipMPI("MPICH2")
@unittest.skipMPI("MPICH1")
class TestIOViewWorld(BaseTestIOView, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


@unittest.skipMPI("msmpi")
@unittest.skipMPI("openmpi")
@unittest.skipMPI("impi", os.name == "nt")
class TestDatarep(unittest.TestCase):
    #
    def testRegister(self):
        def extent_fn(dtype):
            return dtype.extent

        try:
            MPI.Register_datarep(
                "mpi4py-datarep-dummy",
                read_fn=None,
                write_fn=None,
                extent_fn=extent_fn,
            )
        except NotImplementedError:
            self.skipTest("mpi-register-datrep")
        with self.assertRaises(MPI.Exception) as cm:
            MPI.Register_datarep(
                "mpi4py-datarep-dummy",
                read_fn=None,
                write_fn=None,
                extent_fn=extent_fn,
            )
        ierr = cm.exception.Get_error_class()
        self.assertEqual(ierr, MPI.ERR_DUP_DATAREP)


def have_feature():
    case = BaseTestIO()
    case.COMM = TestIOBasicSelf.COMM
    case.prefix = TestIOBasicSelf.prefix
    case.setUp()
    case.tearDown()


try:
    have_feature()
except NotImplementedError:
    unittest.disable(BaseTestIO, "mpi-io")


if __name__ == "__main__":
    unittest.main()



import mpiunittest as unittest

from mpi4py import MPI


class TestCommNull(unittest.TestCase):
    #
    def testConstructor(self):
        comm = MPI.Comm()
        self.assertEqual(comm, MPI.COMM_NULL)
        self.assertIsNot(comm, MPI.COMM_NULL)

        def construct():
            MPI.Comm((1, 2, 3))

        self.assertRaises(TypeError, construct)

    def testConstructorIntra(self):
        comm_null = MPI.Intracomm()
        self.assertEqual(comm_null, MPI.COMM_NULL)
        self.assertIsNot(comm_null, MPI.COMM_NULL)

    def testConstructorInter(self):
        comm_null = MPI.Intercomm()
        self.assertEqual(comm_null, MPI.COMM_NULL)
        self.assertIsNot(comm_null, MPI.COMM_NULL)

    def testGetName(self):
        name = MPI.COMM_NULL.Get_name()
        self.assertEqual(name, "MPI_COMM_NULL")

    def testPickle(self):
        from pickle import dumps, loads

        comm_null = loads(dumps(MPI.COMM_NULL))
        self.assertIs(comm_null, MPI.COMM_NULL)
        comm_null = loads(dumps(MPI.Comm(MPI.COMM_NULL)))
        self.assertIsNot(comm_null, MPI.COMM_NULL)
        self.assertEqual(comm_null, MPI.COMM_NULL)


class BaseTestComm:
    #
    def testConstructor(self):
        comm = MPI.Comm(self.COMM)
        self.assertEqual(comm, self.COMM)
        self.assertIsNot(comm, self.COMM)

    def testSize(self):
        size = self.COMM.Get_size()
        self.assertGreaterEqual(size, 1)

    def testRank(self):
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        self.assertGreaterEqual(rank, 0)
        self.assertLess(rank, size)

    def testGroup(self):
        comm = self.COMM
        group = self.COMM.Get_group()
        self.assertEqual(comm.Get_size(), group.Get_size())
        self.assertEqual(comm.Get_rank(), group.Get_rank())
        group.Free()
        self.assertEqual(group, MPI.GROUP_NULL)

    def testCloneFree(self):
        comm = self.COMM.Clone()
        comm.Free()
        self.assertEqual(comm, MPI.COMM_NULL)

    def testCompare(self):
        results = (MPI.IDENT, MPI.CONGRUENT, MPI.SIMILAR, MPI.UNEQUAL)
        ccmp = MPI.Comm.Compare(self.COMM, MPI.COMM_WORLD)
        self.assertIn(ccmp, results)
        ccmp = MPI.Comm.Compare(self.COMM, self.COMM)
        self.assertEqual(ccmp, MPI.IDENT)
        comm = self.COMM.Dup()
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        comm.Free()
        self.assertEqual(ccmp, MPI.CONGRUENT)

    def testIsInter(self):
        is_inter = self.COMM.Is_inter()
        self.assertIs(type(is_inter), bool)

    def testGetSetName(self):
        try:
            name = self.COMM.Get_name()
            self.COMM.Set_name("mycomm")
            self.assertEqual(self.COMM.Get_name(), "mycomm")
            self.COMM.Set_name(name)
            self.assertEqual(self.COMM.Get_name(), name)
            self.COMM.name = self.COMM.name
        except NotImplementedError:
            self.skipTest("mpi-comm-name")

    def testGetParent(self):
        try:
            parent = MPI.Comm.Get_parent()
            self.assertIsInstance(parent, MPI.Intercomm)
        except NotImplementedError:
            self.skipTest("mpi-comm-get_parent")

    def testDupWithInfo(self):
        info = None
        self.COMM.Dup(info).Free()
        info = MPI.INFO_NULL
        self.COMM.Dup(info).Free()
        self.COMM.Dup_with_info(info).Free()
        info = MPI.Info.Create()
        self.COMM.Dup(info).Free()
        self.COMM.Dup_with_info(info).Free()
        info.Free()

    @unittest.skipMPI("mpich(<=3.1.0)", MPI.Query_thread() > MPI.THREAD_SINGLE)
    def testIDup(self):
        try:
            comm, request = self.COMM.Idup()
        except NotImplementedError:
            self.skipTest("mpi-comm-idup")
        request.Wait()
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        comm.Free()
        self.assertEqual(ccmp, MPI.CONGRUENT)

    @unittest.skipMPI("mpich(<=3.1.0)", MPI.Query_thread() > MPI.THREAD_SINGLE)
    def testIDupWithInfo(self):
        try:
            comm, request = self.COMM.Idup_with_info(MPI.INFO_NULL)
        except NotImplementedError:
            self.skipTest("mpi-comm-idup-info")
        request.Wait()
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        comm.Free()
        self.assertEqual(ccmp, MPI.CONGRUENT)
        #
        new_info = MPI.Info.Create()
        for info in (None, MPI.INFO_NULL, new_info):
            comm, request = self.COMM.Idup(info)
            request.Wait()
            ccmp = MPI.Comm.Compare(self.COMM, comm)
            comm.Free()
            self.assertEqual(ccmp, MPI.CONGRUENT)
        new_info.Free()

    def testGetSetInfo(self):
        info = MPI.Info.Create()
        self.COMM.Set_info(info)
        info.Free()
        info = self.COMM.Get_info()
        self.COMM.Set_info(info)
        info.Free()

    def testCreate(self):
        group = self.COMM.Get_group()
        comm = self.COMM.Create(group)
        ccmp = MPI.Comm.Compare(self.COMM, comm)
        self.assertEqual(ccmp, MPI.CONGRUENT)
        ccmp = self.COMM.Compare(comm)
        self.assertEqual(ccmp, MPI.CONGRUENT)
        comm.Free()
        group.Free()

    @unittest.skipMPI("openmpi(<=1.8.1)")
    def testCreateGroup(self):
        group = self.COMM.Get_group()
        try:
            comm = self.COMM.Create_group(group)
        except NotImplementedError:
            self.assertLess(MPI.Get_version(), (3, 0))
            self.skipTest("mpi-comm-create_group")
        else:
            ccmp = MPI.Comm.Compare(self.COMM, comm)
            comm.Free()
            self.assertEqual(ccmp, MPI.CONGRUENT)
        finally:
            group.Free()

    def testCreateFromGroup(self):
        group = self.COMM.Get_group()
        try:
            comm = MPI.Intracomm.Create_from_group(group)
        except NotImplementedError:
            self.assertLess(MPI.Get_version(), (4, 0))
            self.skipTest("mpi-comm-create_from_group")
        except MPI.Exception as exc:  # openmpi
            UNSUPPORTED = MPI.ERR_UNSUPPORTED_OPERATION
            if exc.Get_error_class() != UNSUPPORTED:
                raise
        else:
            ccmp = MPI.Comm.Compare(self.COMM, comm)
            comm.Free()
            self.assertEqual(ccmp, MPI.CONGRUENT)
        finally:
            group.Free()

    def testSplit(self):
        base = self.COMM
        comm = base.Split(42, 42)
        self.assertEqual(comm.Get_rank(), base.Get_rank())
        self.assertEqual(comm.Get_size(), base.Get_size())
        comm.Free()
        color = base.Get_rank()
        comm = base.Split(color, 42)
        self.assertEqual(comm.Get_rank(), 0)
        self.assertEqual(comm.Get_size(), 1)
        comm.Free()

    @unittest.skipMPI("openmpi(==2.0.0)")
    def testSplitTypeShared(self):
        try:
            MPI.COMM_SELF.Split_type(MPI.COMM_TYPE_SHARED).Free()
        except NotImplementedError:
            self.skipTest("mpi-comm-split_type")
        comm = self.COMM.Split_type(MPI.UNDEFINED)
        self.assertEqual(comm, MPI.COMM_NULL)
        comm = self.COMM.Split_type(MPI.COMM_TYPE_SHARED)
        self.assertNotEqual(comm, MPI.COMM_NULL)
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        if size == 1:
            self.assertEqual(comm.size, 1)
            self.assertEqual(comm.rank, 0)
        comm.Free()
        for root in range(size):
            if rank == root:
                split_type = MPI.COMM_TYPE_SHARED
            else:
                split_type = MPI.UNDEFINED
            comm = self.COMM.Split_type(split_type)
            if rank == root:
                self.assertNotEqual(comm, MPI.COMM_NULL)
                self.assertEqual(comm.size, 1)
                self.assertEqual(comm.rank, 0)
                comm.Free()
            else:
                self.assertEqual(comm, MPI.COMM_NULL)

    def testSplitTypeHWGuided(self):
        try:
            MPI.COMM_SELF.Split_type(MPI.COMM_TYPE_SHARED).Free()
        except NotImplementedError:
            self.skipTest("mpi-comm-split_type")
        if MPI.Get_version() < (4, 0):
            self.skipTest("mpi-comm-split_type-hw_guided")
        split_type = MPI.COMM_TYPE_HW_GUIDED
        #
        comm = self.COMM.Split_type(split_type)
        self.assertEqual(comm, MPI.COMM_NULL)
        comm = self.COMM.Split_type(split_type, info=MPI.INFO_NULL)
        self.assertEqual(comm, MPI.COMM_NULL)
        info = MPI.Info.Create()
        comm = self.COMM.Split_type(split_type, info=info)
        self.assertEqual(comm, MPI.COMM_NULL)
        info.Set("foo", "bar")
        comm = self.COMM.Split_type(split_type, info=info)
        self.assertEqual(comm, MPI.COMM_NULL)
        info.Set("mpi_hw_resource_type", "@dont-thread-on-me@")
        comm = self.COMM.Split_type(split_type, info=info)
        if unittest.is_mpi("impi"):
            comm.free()
        self.assertEqual(comm, MPI.COMM_NULL)
        info.Free()
        #
        if unittest.is_mpi("impi"):
            return
        restype = "mpi_hw_resource_type"
        shmem = "mpi_shared_memory"
        info = MPI.Info.Create()
        info.Set(restype, shmem)
        comm = self.COMM.Split_type(split_type, info=info)
        self.assertNotEqual(comm, MPI.COMM_NULL)
        self.assertEqual(info.Get(restype), shmem)
        comm.Free()
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for root in range(size):
            if rank == root:
                split_type = MPI.COMM_TYPE_HW_GUIDED
            else:
                split_type = MPI.UNDEFINED
            comm = self.COMM.Split_type(split_type, info=info)
            self.assertEqual(info.Get(restype), shmem)
            if rank == root:
                self.assertNotEqual(comm, MPI.COMM_NULL)
                self.assertEqual(comm.size, 1)
                self.assertEqual(comm.rank, 0)
                comm.Free()
            else:
                self.assertEqual(comm, MPI.COMM_NULL)
        info.Free()

    def testSplitTypeResourceGuided(self):
        try:
            MPI.COMM_SELF.Split_type(MPI.COMM_TYPE_SHARED).Free()
        except NotImplementedError:
            self.skipTest("mpi-comm-split_type")
        if MPI.Get_version() < (4, 1):
            self.skipTest("mpi-comm-split_type-resource_guided")
        split_type = MPI.COMM_TYPE_RESOURCE_GUIDED
        #
        comm = self.COMM.Split_type(split_type)
        self.assertEqual(comm, MPI.COMM_NULL)
        comm = self.COMM.Split_type(split_type, info=MPI.INFO_NULL)
        self.assertEqual(comm, MPI.COMM_NULL)
        info = MPI.Info.Create()
        comm = self.COMM.Split_type(split_type, info=info)
        self.assertEqual(comm, MPI.COMM_NULL)
        info.Set("foo", "bar")
        comm = self.COMM.Split_type(split_type, info=info)
        self.assertEqual(comm, MPI.COMM_NULL)
        info.Set("mpi_hw_resource_type", "@dont-thread-on-me@")
        comm = self.COMM.Split_type(split_type, info=info)
        if unittest.is_mpi("impi"):
            comm.free()
        self.assertEqual(comm, MPI.COMM_NULL)
        info.Free()
        #
        restype = "mpi_hw_resource_type"
        shmem = "mpi_shared_memory"
        info = MPI.Info.Create()
        info.Set(restype, shmem)
        comm = self.COMM.Split_type(split_type, info=info)
        self.assertNotEqual(comm, MPI.COMM_NULL)
        self.assertEqual(info.Get(restype), shmem)
        comm.Free()
        size = self.COMM.Get_size()
        rank = self.COMM.Get_rank()
        for root in range(size):
            if rank == root:
                split_type = MPI.COMM_TYPE_RESOURCE_GUIDED
            else:
                split_type = MPI.UNDEFINED
            comm = self.COMM.Split_type(split_type, info=info)
            self.assertEqual(info.Get(restype), shmem)
            if rank == root:
                self.assertNotEqual(comm, MPI.COMM_NULL)
                self.assertEqual(comm.size, 1)
                self.assertEqual(comm.rank, 0)
                comm.Free()
            else:
                self.assertEqual(comm, MPI.COMM_NULL)
        info.Free()

    def testSplitTypeHWUnguided(self):
        try:
            MPI.COMM_SELF.Split_type(MPI.COMM_TYPE_SHARED).Free()
        except NotImplementedError:
            self.skipTest("mpi-comm-split_type")
        if MPI.Get_version() < (4, 0):
            self.skipTest("mpi-comm-split_type-hw_unguided")
        hwcomm = [self.COMM]
        while len(hwcomm) < 32:
            rank = hwcomm[-1].Get_rank()
            info = MPI.Info.Create()
            comm = hwcomm[-1].Split_type(
                MPI.COMM_TYPE_HW_UNGUIDED,
                key=rank,
                info=info,
            )
            if comm != MPI.COMM_NULL:
                self.assertTrue(info.Get("mpi_hw_resource_type"))
                self.assertLess(comm.Get_size(), hwcomm[-1].Get_size())
            info.Free()
            if comm == MPI.COMM_NULL:
                break
            hwcomm.append(comm)
        for comm in hwcomm[1:]:
            comm.Free()

    def testBuffering(self):
        comm = self.COMM.Dup()
        buf = MPI.Alloc_mem((1 << 16) + MPI.BSEND_OVERHEAD)
        try:
            with self.catchNotImplementedError(4, 1):
                comm.Attach_buffer(buf)
            with self.catchNotImplementedError(4, 1):
                comm.Flush_buffer()
            with self.catchNotImplementedError(4, 1):
                comm.Iflush_buffer().Wait()
        finally:
            with self.catchNotImplementedError(4, 1):
                oldbuf = comm.Detach_buffer()
                self.assertEqual(oldbuf.address, buf.address)
                self.assertEqual(oldbuf.nbytes, buf.nbytes)
            MPI.Free_mem(buf)
            with self.catchNotImplementedError(4, 1):
                comm.Attach_buffer(MPI.BUFFER_AUTOMATIC)
                bufauto = comm.Detach_buffer()
                self.assertEqual(bufauto, MPI.BUFFER_AUTOMATIC)
            comm.Free()

    def testPickle(self):
        from pickle import dumps, loads

        COMM = self.COMM
        if COMM in (MPI.COMM_SELF, MPI.COMM_WORLD):
            comm = loads(dumps(COMM))
            self.assertIs(comm, COMM)
            comm = loads(dumps(MPI.Intracomm(COMM)))
            self.assertIsNot(comm, COMM)
            self.assertEqual(comm, COMM)
        else:
            self.assertRaises(ValueError, dumps, COMM)

    def testPyProps(self):
        comm = self.COMM
        self.assertEqual(comm.Get_size(), comm.size)
        self.assertEqual(comm.Get_rank(), comm.rank)
        self.assertEqual(comm.Is_intra(), comm.is_intra)
        self.assertEqual(comm.Is_inter(), comm.is_inter)
        self.assertEqual(comm.Get_topology(), comm.topology)
        #
        group = comm.group
        self.assertEqual(type(group), MPI.Group)
        group.Free()
        #
        info = comm.info
        self.assertIs(type(info), MPI.Info)
        comm.info = info
        info.Free()


class TestCommSelf(BaseTestComm, unittest.TestCase):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF

    def testSize(self):
        size = self.COMM.Get_size()
        self.assertEqual(size, 1)

    def testRank(self):
        rank = self.COMM.Get_rank()
        self.assertEqual(rank, 0)

    @unittest.skipMPI("openmpi(<5.0.6)")
    def testCreateFromGroup(self):
        super().testCreateFromGroup()


class TestCommWorld(BaseTestComm, unittest.TestCase):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD


class TestCommSelfDup(TestCommSelf):
    #
    def setUp(self):
        self.COMM = MPI.COMM_SELF.Dup()

    def tearDown(self):
        self.COMM.Free()


@unittest.skipMPI("openmpi(<1.4.0)", MPI.Query_thread() > MPI.THREAD_SINGLE)
class TestCommWorldDup(TestCommWorld):
    #
    def setUp(self):
        self.COMM = MPI.COMM_WORLD.Dup()

    def tearDown(self):
        self.COMM.Free()


if __name__ == "__main__":
    unittest.main()



import os

import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI


class TestEnviron(unittest.TestCase):
    #
    def testIsInitialized(self):
        flag = MPI.Is_initialized()
        self.assertIs(type(flag), bool)
        self.assertTrue(flag)

    def testIsFinalized(self):
        flag = MPI.Is_finalized()
        self.assertIs(type(flag), bool)
        self.assertFalse(flag)

    def testGetVersion(self):
        version = MPI.Get_version()
        self.assertEqual(len(version), 2)
        major, minor = version
        self.assertIs(type(major), int)
        self.assertIs(type(minor), int)
        self.assertGreaterEqual(major, 1)
        self.assertGreaterEqual(minor, 0)

    def testGetLibraryVersion(self):
        version = MPI.Get_library_version()
        self.assertIsInstance(version, str)
        self.assertGreater(len(version), 0)

    def testGetProcessorName(self):
        procname = MPI.Get_processor_name()
        self.assertIsInstance(procname, str)

    def testGetHWResourceInfo(self):
        with self.catchNotImplementedError(4, 1):
            info = MPI.Get_hw_resource_info()
            self.assertIsInstance(info, MPI.Info)

    def testWTime(self):
        time1 = MPI.Wtime()
        self.assertIs(type(time1), float)
        time2 = MPI.Wtime()
        self.assertIs(type(time2), float)
        self.assertGreaterEqual(time2, time1)

    @unittest.skipMPI("impi", os.name == "nt")
    def testWTick(self):
        tick = MPI.Wtick()
        self.assertIs(type(tick), float)
        self.assertGreater(tick, 0.0)

    def testPControl(self):
        for level in (2, 1, 0):
            MPI.Pcontrol(level)
        MPI.Pcontrol(1)


class TestWorldAttrs(unittest.TestCase):
    #
    def testWTimeIsGlobal(self):
        wtg = MPI.COMM_WORLD.Get_attr(MPI.WTIME_IS_GLOBAL)
        if wtg is not None:
            self.assertIn(wtg, (True, False))

    def testIOProcessor(self):
        size = MPI.COMM_WORLD.Get_size()
        vals = list(range(size))
        vals += [MPI.UNDEFINED, MPI.ANY_SOURCE, MPI.PROC_NULL]
        ioproc = MPI.COMM_WORLD.Get_attr(MPI.IO)
        if ioproc is not None:
            self.assertIn(ioproc, vals)

    @unittest.skipIf(MPI.APPNUM == MPI.KEYVAL_INVALID, "mpi-appnum")
    def testAppNum(self):
        appnum = MPI.COMM_WORLD.Get_attr(MPI.APPNUM)
        if appnum is not None:
            self.assertTrue(appnum == MPI.UNDEFINED or appnum >= 0)

    @unittest.skipMPI("mpich(==4.3.1)", not testutil.has_mpi_appnum())
    @unittest.skipMPI("mpich(<4.1.0)", not testutil.has_mpi_appnum())
    @unittest.skipMPI("mvapich", not testutil.has_mpi_appnum())
    @unittest.skipMPI("MPICH2", not testutil.has_mpi_appnum())
    @unittest.skipIf(
        MPI.UNIVERSE_SIZE == MPI.KEYVAL_INVALID, "mpi-universe-size"
    )
    def testUniverseSize(self):
        univsz = MPI.COMM_WORLD.Get_attr(MPI.UNIVERSE_SIZE)
        if univsz is not None:
            self.assertTrue(univsz == MPI.UNDEFINED or univsz >= 0)

    @unittest.skipIf(
        MPI.LASTUSEDCODE == MPI.KEYVAL_INVALID, "mpi-lastusedcode"
    )
    def testLastUsedCode(self):
        lastuc = MPI.COMM_WORLD.Get_attr(MPI.LASTUSEDCODE)
        self.assertGreaterEqual(lastuc, 0)


if __name__ == "__main__":
    unittest.main()



import sys
from typing import (
    Callable,
    Literal,
)

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ..MPI import (
    COMM_SELF,
    INFO_NULL,
    Info,
    Intracomm,
)
from ..typing import T

__all__: list[str] = [
    "Sequential",
    "Counter",
    "Mutex",
    "Condition",
    "Semaphore",
]

class Sequential:
    comm: Intracomm
    tag: int
    def __init__(
        self,
        comm: Intracomm,
        tag: int = 0,
    ) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *exc: object) -> None: ...
    def begin(self) -> None: ...
    def end(self) -> None: ...

class Counter:
    def __init__(
        self,
        start: int = 0,
        step: int = 1,
        *,
        typecode: str = "i",
        comm: Intracomm = COMM_SELF,
        info: Info = INFO_NULL,
        root: int = 0,
    ) -> None: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> int: ...
    def next(self, incr: int | None = None) -> int: ...
    def free(self) -> None: ...

class Mutex:
    def __init__(
        self,
        *,
        recursive: bool = False,
        comm: Intracomm = COMM_SELF,
        info: Info = INFO_NULL,
    ) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *exc: object) -> None: ...
    def acquire(self, blocking: bool = True) -> bool: ...
    def release(self) -> None: ...
    def locked(self) -> bool: ...
    def count(self) -> int: ...
    def free(self) -> None: ...

class Condition:
    def __init__(
        self,
        mutex: Mutex | None = None,
        *,
        recursive: bool = True,
        comm: Intracomm = COMM_SELF,
        info: Info = INFO_NULL,
    ) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *exc: object) -> None: ...
    def acquire(self, blocking: bool = True) -> bool: ...
    def release(self) -> None: ...
    def locked(self) -> bool: ...
    def wait(self) -> Literal[True]: ...
    def wait_for(self, predicate: Callable[[], T]) -> T: ...
    def notify(self, n: int = 1) -> int: ...
    def notify_all(self) -> int: ...
    def free(self) -> None: ...

class Semaphore:
    def __init__(
        self,
        value: int = 1,
        *,
        bounded: bool = True,
        comm: Intracomm = COMM_SELF,
        info: Info = INFO_NULL,
    ) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, *exc: object) -> None: ...
    def acquire(self, blocking: bool = True) -> bool: ...
    def release(self, n: int = 1) -> None: ...
    def free(self) -> None: ...
